@@ -7,6 +7,7 @@
 // - Detlev Vendt (detlev.vendt@brillit.de)
 // - Petr Supina (psup@centrum.cz)
 // - Carsten Klein (c.klein@datagis.com)
+// - Mihail Naydenov (mnaydenov@users.sourceforge.net)
 //
 // This file is part of FreeImage 3
 //
@@ -28,9 +29,9 @@
 #endif 
 
 #include <stdlib.h>
-#if defined(_WIN32) || defined(_WIN64)
+#if defined(_WIN32) || defined(_WIN64) || defined(__MINGW32__)
 #include <malloc.h>
-#endif // _WIN32 || _WIN64
+#endif // _WIN32 || _WIN64 || __MINGW32__
 
 #include "FreeImage.h"
 #include "FreeImageIO.h"
@@ -59,7 +60,7 @@ FI_STRUCT (METADATAHEADER) {
 // ----------------------------------------------------------
 
 FI_STRUCT (FREEIMAGEHEADER) {
-    FREE_IMAGE_TYPE type;		// data type - bitmap, array of long, double, complex, etc
+	FREE_IMAGE_TYPE type;		// data type - bitmap, array of long, double, complex, etc
 
 	unsigned red_mask;			// bit layout of the red components
 	unsigned green_mask;		// bit layout of the green components
@@ -67,13 +68,17 @@ FI_STRUCT (FREEIMAGEHEADER) {
 
 	RGBQUAD bkgnd_color;		// background color used for RGB transparency
 
-	BOOL transparent;			 // why another table? for easy transparency table retrieval!
-	int  transparency_count;	 // transparency could be stored in the palette, which is better
-	BYTE transparent_table[256]; // overall, but it requires quite some changes and it will render
-								 // FreeImage_GetTransparencyTable obsolete in its current form;
-	FIICCPROFILE iccProfile;	 // space to hold ICC profile
+	BOOL transparent;			// why another table? for easy transparency table retrieval!
+	int  transparency_count;	// transparency could be stored in the palette, which is better
+	BYTE transparent_table[256];// overall, but it requires quite some changes and it will render
+								// FreeImage_GetTransparencyTable obsolete in its current form;
+	FIICCPROFILE iccProfile;	// space to hold ICC profile
 
-	METADATAMAP *metadata;		 // contains a list of metadata models attached to the bitmap
+	METADATAMAP *metadata;		// contains a list of metadata models attached to the bitmap
+
+	BOOL has_pixels;			// FALSE if the FIBITMAP only contains the header and no pixel data
+
+	FIBITMAP *thumbnail;		// optionally contains a thumbnail attached to the bitmap
 
 	//BYTE filler[1];			 // fill to 32-bit alignment
 };
@@ -82,7 +87,7 @@ FI_STRUCT (FREEIMAGEHEADER) {
 //  Memory allocation on a specified alignment boundary
 // ----------------------------------------------------------
 
-#if defined(_WIN32) || defined(_WIN64)
+#if (defined(_WIN32) || defined(_WIN64)) && !defined(__MINGW32__)
 
 void* FreeImage_Aligned_Malloc(size_t amount, size_t alignment) {
 	assert(alignment == FIBITMAP_ALIGNMENT);
@@ -93,6 +98,17 @@ void FreeImage_Aligned_Free(void* mem) {
 	_aligned_free(mem);
 }
 
+#elif defined (__MINGW32__)
+
+void* FreeImage_Aligned_Malloc(size_t amount, size_t alignment) {
+	assert(alignment == FIBITMAP_ALIGNMENT);
+	return __mingw_aligned_malloc (amount, alignment);
+}
+
+void FreeImage_Aligned_Free(void* mem) {
+	__mingw_aligned_free (mem);
+}
+
 #else
 
 void* FreeImage_Aligned_Malloc(size_t amount, size_t alignment) {
@@ -133,32 +149,59 @@ void FreeImage_Aligned_Free(void* mem) {
 /**
 Calculate the size of a FreeImage image. 
 Align the palette and the pixels on a FIBITMAP_ALIGNMENT bytes alignment boundary.
+
+@param header_only If TRUE, calculate a 'header only' FIBITMAP size, otherwise calculate a full FIBITMAP size
+@param width
+@param height
+@param bpp
+@see FreeImage_AllocateHeaderT
 */
-static unsigned 
-FreeImage_GetImageSize(int width, int height, int bpp) {
-	unsigned dib_size = sizeof(FREEIMAGEHEADER); 
+static size_t 
+FreeImage_GetImageSizeHeader(BOOL header_only, unsigned width, unsigned height, unsigned bpp) {
+	size_t dib_size = sizeof(FREEIMAGEHEADER); 
 	dib_size += (dib_size % FIBITMAP_ALIGNMENT ? FIBITMAP_ALIGNMENT - dib_size % FIBITMAP_ALIGNMENT : 0);  
 	dib_size += FIBITMAP_ALIGNMENT - sizeof(BITMAPINFOHEADER) % FIBITMAP_ALIGNMENT; 
 	dib_size += sizeof(BITMAPINFOHEADER);  
 	// palette is aligned on a 16 bytes boundary
 	dib_size += sizeof(RGBQUAD) * CalculateUsedPaletteEntries(bpp);  
 	dib_size += (dib_size % FIBITMAP_ALIGNMENT ? FIBITMAP_ALIGNMENT - dib_size % FIBITMAP_ALIGNMENT : 0);  
-	// pixels are aligned on a 16 bytes boundary
-	dib_size += CalculatePitch(CalculateLine(width, bpp)) * height; 
+	if(!header_only) {
+		const size_t header_size = dib_size;
+
+		// pixels are aligned on a 16 bytes boundary
+		dib_size += CalculatePitch(CalculateLine(width, bpp)) * height; 
+
+		// check for possible malloc overflow using a KISS integer overflow detection mechanism
+		{
+			/*
+			The following constant take into account the additionnal memory used by 
+			aligned malloc functions as well as debug malloc functions. 
+			It is supposed here that using a (8 * FIBITMAP_ALIGNMENT) risk margin will be enough
+			for the target compiler. 
+			*/
+			const double FIBITMAP_MAX_MEMORY = (double)((size_t)-1) - 8 * FIBITMAP_ALIGNMENT;
+			const double dPitch = floor( ((double)bpp * width + 31.0) / 32.0 ) * 4.0;
+			const double dImageSize = (double)header_size + dPitch * height;
+			if(dImageSize != (double)dib_size) {
+				// here, we are sure to encounter a malloc overflow: try to avoid it ...
+				return 0;
+			}
+			if(dImageSize > FIBITMAP_MAX_MEMORY) {
+				// avoid possible overflow inside C allocation functions
+				return 0;
+			}
+		}
+	}
 
 	return dib_size;
 }
 
 FIBITMAP * DLL_CALLCONV
-FreeImage_Allocate(int width, int height, int bpp, unsigned red_mask, unsigned green_mask, unsigned blue_mask) {
-	return FreeImage_AllocateT(FIT_BITMAP, width, height, bpp, red_mask, green_mask, blue_mask);
-}
-
-FIBITMAP * DLL_CALLCONV
-FreeImage_AllocateT(FREE_IMAGE_TYPE type, int width, int height, int bpp, unsigned red_mask, unsigned green_mask, unsigned blue_mask) {
+FreeImage_AllocateHeaderT(BOOL header_only, FREE_IMAGE_TYPE type, int width, int height, int bpp, unsigned red_mask, unsigned green_mask, unsigned blue_mask) {
 	FIBITMAP *bitmap = (FIBITMAP *)malloc(sizeof(FIBITMAP));
 
 	if (bitmap != NULL) {
+		width = abs(width);
 		height = abs(height);
 
 		// check pixel bit depth
@@ -184,10 +227,10 @@ FreeImage_AllocateT(FREE_IMAGE_TYPE type, int width, int height, int bpp, unsign
 				bpp = 8 * sizeof(short);
 				break;
 			case FIT_UINT32:
-				bpp = 8 * sizeof(unsigned long);
+				bpp = 8 * sizeof(DWORD);
 				break;
 			case FIT_INT32:
-				bpp = 8 * sizeof(long);
+				bpp = 8 * sizeof(LONG);
 				break;
 			case FIT_FLOAT:
 				bpp = 8 * sizeof(float);
@@ -220,7 +263,13 @@ FreeImage_AllocateT(FREE_IMAGE_TYPE type, int width, int height, int bpp, unsign
 		// palette is aligned on a 16 bytes boundary
 		// pixels are aligned on a 16 bytes boundary
 
-		unsigned dib_size = FreeImage_GetImageSize(width, height, bpp); 
+		size_t dib_size = FreeImage_GetImageSizeHeader(header_only, width, height, bpp); 
+
+		if(dib_size == 0) {
+			// memory allocation will fail (probably a malloc overflow)
+			free(bitmap);
+			return NULL;
+		}
 
 		bitmap->data = (BYTE *)FreeImage_Aligned_Malloc(dib_size * sizeof(BYTE), FIBITMAP_ALIGNMENT);
 
@@ -242,6 +291,8 @@ FreeImage_AllocateT(FREE_IMAGE_TYPE type, int width, int height, int bpp, unsign
 			fih->transparency_count = 0;
 			memset(fih->transparent_table, 0xff, 256);
 
+			fih->has_pixels = header_only ? FALSE : TRUE;
+
 			// initialize FIICCPROFILE link
 
 			FIICCPROFILE *iccProfile = FreeImage_GetICCProfile(bitmap);
@@ -251,7 +302,11 @@ FreeImage_AllocateT(FREE_IMAGE_TYPE type, int width, int height, int bpp, unsign
 
 			// initialize metadata models list
 
-			fih->metadata = new METADATAMAP;
+			fih->metadata = new(std::nothrow) METADATAMAP;
+
+			// initialize attached thumbnail
+
+			fih->thumbnail = NULL;
 
 			// write out the BITMAPINFOHEADER
 
@@ -267,6 +322,16 @@ FreeImage_AllocateT(FREE_IMAGE_TYPE type, int width, int height, int bpp, unsign
 			bih->biXPelsPerMeter	= 2835;	// 72 dpi
 			bih->biYPelsPerMeter	= 2835;	// 72 dpi
 
+			if(bpp == 8) {
+				// build a default greyscale palette (very useful for image processing)
+				RGBQUAD *pal = FreeImage_GetPalette(bitmap);
+				for(int i = 0; i < 256; i++) {
+					pal[i].rgbRed	= (BYTE)i;
+					pal[i].rgbGreen = (BYTE)i;
+					pal[i].rgbBlue	= (BYTE)i;
+				}
+			}
+
 			return bitmap;
 		}
 
@@ -276,6 +341,21 @@ FreeImage_AllocateT(FREE_IMAGE_TYPE type, int width, int height, int bpp, unsign
 	return NULL;
 }
 
+FIBITMAP * DLL_CALLCONV
+FreeImage_AllocateHeader(BOOL header_only, int width, int height, int bpp, unsigned red_mask, unsigned green_mask, unsigned blue_mask) {
+	return FreeImage_AllocateHeaderT(header_only, FIT_BITMAP, width, height, bpp, red_mask, green_mask, blue_mask);
+}
+
+FIBITMAP * DLL_CALLCONV
+FreeImage_Allocate(int width, int height, int bpp, unsigned red_mask, unsigned green_mask, unsigned blue_mask) {
+	return FreeImage_AllocateHeaderT(FALSE, FIT_BITMAP, width, height, bpp, red_mask, green_mask, blue_mask);
+}
+
+FIBITMAP * DLL_CALLCONV
+FreeImage_AllocateT(FREE_IMAGE_TYPE type, int width, int height, int bpp, unsigned red_mask, unsigned green_mask, unsigned blue_mask) {
+	return FreeImage_AllocateHeaderT(FALSE, type, width, height, bpp, red_mask, green_mask, blue_mask);
+}
+
 void DLL_CALLCONV
 FreeImage_Unload(FIBITMAP *dib) {
 	if (NULL != dib) {	
@@ -302,6 +382,9 @@ FreeImage_Unload(FIBITMAP *dib) {
 
 			delete metadata;
 
+			// delete embedded thumbnail
+			FreeImage_Unload(FreeImage_GetThumbnail(dib));
+
 			// delete bitmap ...
 			FreeImage_Aligned_Free(dib->data);
 		}
@@ -319,8 +402,11 @@ FreeImage_Clone(FIBITMAP *dib) {
 	unsigned height = FreeImage_GetHeight(dib);
 	unsigned bpp    = FreeImage_GetBPP(dib);
 	
+	// check for pixel availability ...
+	BOOL header_only = FreeImage_HasPixels(dib) ? FALSE : TRUE;
+
 	// allocate a new dib
-	FIBITMAP *new_dib = FreeImage_AllocateT(FreeImage_GetImageType(dib), width, height, bpp, 
+	FIBITMAP *new_dib = FreeImage_AllocateHeaderT(header_only, FreeImage_GetImageType(dib), width, height, bpp, 
 			FreeImage_GetRedMask(dib), FreeImage_GetGreenMask(dib), FreeImage_GetBlueMask(dib));
 
 	if (new_dib) {
@@ -337,7 +423,7 @@ FreeImage_Clone(FIBITMAP *dib) {
 		// palette is aligned on a 16 bytes boundary
 		// pixels are aligned on a 16 bytes boundary
 
-		unsigned dib_size = FreeImage_GetImageSize(width, height, bpp); 
+		size_t dib_size = FreeImage_GetImageSizeHeader(header_only, width, height, bpp); 
 
 		// copy the bitmap + internal pointers (remember to restore new_dib internal pointers later)
 		memcpy(new_dib->data, dib->data, dib_size);
@@ -348,6 +434,9 @@ FreeImage_Clone(FIBITMAP *dib) {
 		// restore metadata link for new_dib
 		((FREEIMAGEHEADER *)new_dib->data)->metadata = dst_metadata;
 
+		// reset thumbnail link for new_dib
+		((FREEIMAGEHEADER *)new_dib->data)->thumbnail = NULL;
+
 		// copy possible ICC profile
 		FreeImage_CreateICCProfile(new_dib, src_iccProfile->data, src_iccProfile->size);
 		dst_iccProfile->flags = src_iccProfile->flags;
@@ -359,22 +448,27 @@ FreeImage_Clone(FIBITMAP *dib) {
 
 			if(src_tagmap) {
 				// create a metadata model
-				TAGMAP *dst_tagmap = new TAGMAP();
+				TAGMAP *dst_tagmap = new(std::nothrow) TAGMAP();
 
-				// fill the model
-				for(TAGMAP::iterator j = src_tagmap->begin(); j != src_tagmap->end(); j++) {
-					std::string dst_key = (*j).first;
-					FITAG *dst_tag = FreeImage_CloneTag( (*j).second );
+				if(dst_tagmap) {
+					// fill the model
+					for(TAGMAP::iterator j = src_tagmap->begin(); j != src_tagmap->end(); j++) {
+						std::string dst_key = (*j).first;
+						FITAG *dst_tag = FreeImage_CloneTag( (*j).second );
 
-					// assign key and tag value
-					(*dst_tagmap)[dst_key] = dst_tag;
+						// assign key and tag value
+						(*dst_tagmap)[dst_key] = dst_tag;
+					}
+
+					// assign model and tagmap
+					(*dst_metadata)[model] = dst_tagmap;
 				}
-
-				// assign model and tagmap
-				(*dst_metadata)[model] = dst_tagmap;
 			}
 		}
 
+		// copy the thumbnail
+		FreeImage_SetThumbnail(new_dib, FreeImage_GetThumbnail(dib));
+
 		return new_dib;
 	}
 
@@ -383,11 +477,34 @@ FreeImage_Clone(FIBITMAP *dib) {
 
 // ----------------------------------------------------------
 
+FIBITMAP* DLL_CALLCONV
+FreeImage_GetThumbnail(FIBITMAP *dib) {
+	return (dib != NULL) ? ((FREEIMAGEHEADER *)dib->data)->thumbnail : NULL;
+}
+
+BOOL DLL_CALLCONV
+FreeImage_SetThumbnail(FIBITMAP *dib, FIBITMAP *thumbnail) {
+	if(dib == NULL) {
+		return FALSE;
+	}
+	FIBITMAP *currentThumbnail = ((FREEIMAGEHEADER *)dib->data)->thumbnail;
+	if(currentThumbnail == thumbnail) {
+		return TRUE;
+	}
+	FreeImage_Unload(currentThumbnail);
+
+	((FREEIMAGEHEADER *)dib->data)->thumbnail = FreeImage_HasPixels(thumbnail) ? FreeImage_Clone(thumbnail) : NULL;
+
+	return TRUE;
+}
+
+// ----------------------------------------------------------
+
 FREE_IMAGE_COLOR_TYPE DLL_CALLCONV
 FreeImage_GetColorType(FIBITMAP *dib) {
 	RGBQUAD *rgb;
 
-	FREE_IMAGE_TYPE image_type = FreeImage_GetImageType(dib);
+	const FREE_IMAGE_TYPE image_type = FreeImage_GetImageType(dib);
 
 	// special bitmap type
 	if(image_type != FIT_BITMAP) {
@@ -486,6 +603,13 @@ FreeImage_GetImageType(FIBITMAP *dib) {
 
 // ----------------------------------------------------------
 
+BOOL DLL_CALLCONV 
+FreeImage_HasPixels(FIBITMAP *dib) {
+	return (dib != NULL) ? ((FREEIMAGEHEADER *)dib->data)->has_pixels : FALSE;
+}
+
+// ----------------------------------------------------------
+
 unsigned DLL_CALLCONV
 FreeImage_GetRedMask(FIBITMAP *dib) {
 	return dib ? ((FREEIMAGEHEADER *)dib->data)->red_mask : 0;
@@ -567,12 +691,22 @@ FreeImage_SetBackgroundColor(FIBITMAP *dib, RGBQUAD *bkcolor) {
 BOOL DLL_CALLCONV
 FreeImage_IsTransparent(FIBITMAP *dib) {
 	if(dib) {
-		if(FreeImage_GetBPP(dib) == 32) {
-			if(FreeImage_GetColorType(dib) == FIC_RGBALPHA) {
+		FREE_IMAGE_TYPE image_type = FreeImage_GetImageType(dib);
+		switch(image_type) {
+			case FIT_BITMAP:
+				if(FreeImage_GetBPP(dib) == 32) {
+					if(FreeImage_GetColorType(dib) == FIC_RGBALPHA) {
+						return TRUE;
+					}
+				} else {
+					return ((FREEIMAGEHEADER *)dib->data)->transparent ? TRUE : FALSE;
+				}
+				break;
+			case FIT_RGBA16:
+			case FIT_RGBAF:
 				return TRUE;
-			}
-		} else {
-			return ((FREEIMAGEHEADER *)dib->data)->transparent ? TRUE : FALSE;
+			default:
+				break;
 		}
 	}
 	return FALSE;
@@ -602,6 +736,7 @@ FreeImage_GetTransparencyCount(FIBITMAP *dib) {
 void DLL_CALLCONV
 FreeImage_SetTransparencyTable(FIBITMAP *dib, BYTE *table, int count) {
 	if (dib) {
+		count = MIN(count, 256);
 		if (FreeImage_GetBPP(dib) <= 8) {
 			((FREEIMAGEHEADER *)dib->data)->transparent = TRUE;
 			((FREEIMAGEHEADER *)dib->data)->transparency_count = count;
@@ -640,9 +775,9 @@ FreeImage_SetTransparentIndex(FIBITMAP *dib, int index) {
 	if (dib) {
 		int count = FreeImage_GetColorsUsed(dib);
 		if (count) {
-			BYTE *new_tt = (BYTE *)malloc((count - 1) * sizeof(BYTE));
+			BYTE *new_tt = (BYTE *)malloc(count * sizeof(BYTE));
 			memset(new_tt, 0xFF, count);
-			if ((index >= 0) && (index <= count)) {
+			if ((index >= 0) && (index < count)) {
 				new_tt[index] = 0x00;
 			}
 			FreeImage_SetTransparencyTable(dib, new_tt, count);
@@ -756,7 +891,7 @@ FreeImage_GetPalette(FIBITMAP *dib) {
 
 unsigned DLL_CALLCONV
 FreeImage_GetDotsPerMeterX(FIBITMAP *dib) {
-	return FreeImage_GetInfoHeader(dib)->biXPelsPerMeter;
+	return (dib) ? FreeImage_GetInfoHeader(dib)->biXPelsPerMeter : 0;
 }
 
 unsigned DLL_CALLCONV
@@ -803,7 +938,10 @@ FreeImage_FindFirstMetadata(FREE_IMAGE_MDMODEL model, FIBITMAP *dib, FITAG **tag
 
 	// get the metadata model
 	METADATAMAP *metadata = ((FREEIMAGEHEADER *)dib->data)->metadata;
-	TAGMAP *tagmap = (*metadata)[model];
+	TAGMAP *tagmap = NULL;
+	if( (*metadata).find(model) != (*metadata).end() ) {
+		tagmap = (*metadata)[model];
+	}
 	if(tagmap) {
 		// allocate a handle
 		FIMETADATA 	*handle = (FIMETADATA *)malloc(sizeof(FIMETADATA));
@@ -845,7 +983,7 @@ FreeImage_FindNextMetadata(FIMETADATA *mdhandle, FITAG **tag) {
 	TAGMAP *tagmap = mdh->tagmap;
 
 	int current_pos = mdh->pos;
-	int mapsize     = tagmap->size();
+	int mapsize     = (int)tagmap->size();
 
 	if(current_pos < mapsize) {
 		// get the tag element at position pos
@@ -876,6 +1014,57 @@ FreeImage_FindCloseMetadata(FIMETADATA *mdhandle) {
 	}
 }
 
+
+// ----------------------------------------------------------
+
+BOOL DLL_CALLCONV
+FreeImage_CloneMetadata(FIBITMAP *dst, FIBITMAP *src) {
+	if(!src || !dst) return FALSE;
+
+	// get metadata links
+	METADATAMAP *src_metadata = ((FREEIMAGEHEADER *)src->data)->metadata;
+	METADATAMAP *dst_metadata = ((FREEIMAGEHEADER *)dst->data)->metadata;
+
+	// copy metadata models, *except* the FIMD_ANIMATION model
+	for(METADATAMAP::iterator i = (*src_metadata).begin(); i != (*src_metadata).end(); i++) {
+		int model = (*i).first;
+		if(model == (int)FIMD_ANIMATION) {
+			continue;
+		}
+		TAGMAP *src_tagmap = (*i).second;
+
+		if(src_tagmap) {
+			if( dst_metadata->find(model) != dst_metadata->end() ) {
+				// destroy dst model
+				FreeImage_SetMetadata((FREE_IMAGE_MDMODEL)model, dst, NULL, NULL);
+			}
+
+			// create a metadata model
+			TAGMAP *dst_tagmap = new(std::nothrow) TAGMAP();
+
+			if(dst_tagmap) {
+				// fill the model
+				for(TAGMAP::iterator j = src_tagmap->begin(); j != src_tagmap->end(); j++) {
+					std::string dst_key = (*j).first;
+					FITAG *dst_tag = FreeImage_CloneTag( (*j).second );
+
+					// assign key and tag value
+					(*dst_tagmap)[dst_key] = dst_tag;
+				}
+
+				// assign model and tagmap
+				(*dst_metadata)[model] = dst_tagmap;
+			}
+		}
+	}
+
+	// clone resolution 
+	FreeImage_SetDotsPerMeterX(dst, FreeImage_GetDotsPerMeterX(src)); 
+	FreeImage_SetDotsPerMeterY(dst, FreeImage_GetDotsPerMeterY(src)); 
+
+	return TRUE;
+}
+
 // ----------------------------------------------------------
 
 BOOL DLL_CALLCONV 
@@ -887,13 +1076,16 @@ FreeImage_SetMetadata(FREE_IMAGE_MDMODEL model, FIBITMAP *dib, const char *key,
 
 	// get the metadata model
 	METADATAMAP *metadata = ((FREEIMAGEHEADER *)dib->data)->metadata;
-	tagmap = (*metadata)[model];
+	METADATAMAP::iterator model_iterator = metadata->find(model);
+	if (model_iterator != metadata->end()) {
+		tagmap = model_iterator->second;
+	}
 
 	if(key != NULL) {
 
 		if(!tagmap) {
 			// this model, doesn't exist: create it 
-			tagmap = new TAGMAP();
+			tagmap = new(std::nothrow) TAGMAP();
 			(*metadata)[model] = tagmap;
 		}
 		
@@ -905,7 +1097,7 @@ FreeImage_SetMetadata(FREE_IMAGE_MDMODEL model, FIBITMAP *dib, const char *key,
 				// set the tag key
 				FreeImage_SetTagKey(tag, key);
 			}
-			if(FreeImage_GetTagCount(tag) * FreeImage_TagDataWidth((WORD)FreeImage_GetTagType(tag)) != FreeImage_GetTagLength(tag)) {
+			if(FreeImage_GetTagCount(tag) * FreeImage_TagDataWidth(FreeImage_GetTagType(tag)) != FreeImage_GetTagLength(tag)) {
 				FreeImage_OutputMessageProc(FIF_UNKNOWN, "Invalid data count for tag '%s'", key);
 				return FALSE;
 			}
@@ -916,9 +1108,11 @@ FreeImage_SetMetadata(FREE_IMAGE_MDMODEL model, FIBITMAP *dib, const char *key,
 				case FIMD_IPTC:
 				{
 					int id = tag_lib.getTagID(TagLib::IPTC, key);
+					/*
 					if(id == -1) {
 						FreeImage_OutputMessageProc(FIF_UNKNOWN, "IPTC: Invalid key '%s'", key);
 					}
+					*/
 					FreeImage_SetTagID(tag, (WORD)id);
 				}
 				break;
@@ -955,7 +1149,7 @@ FreeImage_SetMetadata(FREE_IMAGE_MDMODEL model, FIBITMAP *dib, const char *key,
 			}
 
 			delete tagmap;
-			(*metadata)[model] = NULL;
+			metadata->erase(model_iterator);
 		}
 	}
 
@@ -964,7 +1158,7 @@ FreeImage_SetMetadata(FREE_IMAGE_MDMODEL model, FIBITMAP *dib, const char *key,
 
 BOOL DLL_CALLCONV 
 FreeImage_GetMetadata(FREE_IMAGE_MDMODEL model, FIBITMAP *dib, const char *key, FITAG **tag) {
-	if(!dib || !key) 
+	if(!dib || !key || !tag) 
 		return FALSE;
 
 	TAGMAP *tagmap = NULL;
@@ -973,15 +1167,18 @@ FreeImage_GetMetadata(FREE_IMAGE_MDMODEL model, FIBITMAP *dib, const char *key,
 	// get the metadata model
 	METADATAMAP *metadata = ((FREEIMAGEHEADER *)dib->data)->metadata;
 	if(!(*metadata).empty()) {
-		tagmap = (*metadata)[model];
-		if(!tagmap) {
-			// this model, doesn't exist: return
-			return FALSE;
+		METADATAMAP::iterator model_iterator = metadata->find(model);
+		if (model_iterator != metadata->end() ) {
+			// this model exists : try to get the requested tag
+			tagmap = model_iterator->second;
+			TAGMAP::iterator tag_iterator = tagmap->find(key);
+			if (tag_iterator != tagmap->end() ) {
+				// get the requested tag
+				*tag = tag_iterator->second;
+			} 
 		}
-
-		// get the requested tag
-		*tag = (*tagmap)[key];
 	}
+
 	return (*tag != NULL) ? TRUE : FALSE;
 }
 
@@ -996,14 +1193,16 @@ FreeImage_GetMetadataCount(FREE_IMAGE_MDMODEL model, FIBITMAP *dib) {
 
 	// get the metadata model
 	METADATAMAP *metadata = ((FREEIMAGEHEADER *)dib->data)->metadata;
-	tagmap = (*metadata)[model];
+	if( (*metadata).find(model) != (*metadata).end() ) {
+		tagmap = (*metadata)[model];
+	}
 	if(!tagmap) {
 		// this model, doesn't exist: return
 		return 0;
 	}
 
 	// get the tag count
-	return tagmap->size();
+	return (unsigned)tagmap->size();
 }
 
 // ----------------------------------------------------------
