@@ -9,6 +9,7 @@
 // - Herv Drolon (drolon@infonie.fr)
 // - Vadim Alexandrov (vadimalexandrov@users.sourceforge.net
 // - Martin Dyring-Andersen (mda@spamfighter.com)
+// - Volodymyr Goncharov (volodymyr.goncharov@gmail.com)
 //
 // This file is part of FreeImage 3
 //
@@ -46,6 +47,7 @@ struct BlockTypeS {
 
 	BlockTypeS(BlockType type) : m_type(type) {
 	}
+	virtual ~BlockTypeS() {}
 };
 
 struct BlockContinueus : public BlockTypeS {
@@ -91,6 +93,24 @@ FI_STRUCT (MULTIBITMAPHEADER) {
 	int load_flags;
 };
 
+// =====================================================================
+// Helper functions
+// =====================================================================
+
+inline void
+ReplaceExtension(std::string& dst_filename, const std::string& src_filename, const std::string& dst_extension) {
+	size_t lastDot = src_filename.find_last_of('.');
+	if (lastDot == std::string::npos) {
+		dst_filename = src_filename;
+		dst_filename += ".";
+		dst_filename += dst_extension;
+	}
+	else {
+		dst_filename = src_filename.substr(0, lastDot + 1);
+		dst_filename += dst_extension;
+	}
+}
+
 // =====================================================================
 // Internal Multipage functions
 // =====================================================================
@@ -212,329 +232,314 @@ FreeImage_InternalGetPageCount(FIMULTIBITMAP *bitmap) {
 // Multipage functions
 // =====================================================================
 
-FIMULTIBITMAP * DLL_CALLCONV
-FreeImage_LoadMultiBitmapFromMemory(FREE_IMAGE_FORMAT fif, FIMEMORY *stream, int flags) {
-	// retrieve the plugin list to find the node belonging to this plugin
-
-	PluginList *list = FreeImage_GetPluginList();
-
-	if (list) {
-		PluginNode *node = list->FindNodeFromFIF(fif);
-
-		if (node) {
-			FreeImageIO *io = new FreeImageIO;
-
-			if (io) {
-				SetMemoryIO(io);
-
-				FIMULTIBITMAP *bitmap = new FIMULTIBITMAP;
-
-				if (bitmap) {
-					MULTIBITMAPHEADER *header = new MULTIBITMAPHEADER;
-
-					header->m_filename = NULL;
-					header->node = node;
-					header->fif = fif;
-					header->io = io;
-					header->handle = (fi_handle)stream;						
-					header->changed = FALSE;						
-					header->read_only = TRUE;
-					header->m_cachefile = NULL;
-					header->cache_fif = fif;
-					header->load_flags = flags;
-
-					if (header) {
-						// store the MULTIBITMAPHEADER in the surrounding FIMULTIBITMAP structure
-
-						bitmap->data = header;
-
-						// cache the page count
-
-						header->page_count = FreeImage_InternalGetPageCount(bitmap);
-
-						// allocate a continueus block to describe the bitmap
-
-						header->m_blocks.push_back((BlockTypeS *)new BlockContinueus(0, header->page_count - 1));
-
-						// set up the cache
-
-						return bitmap;
-					}
-					
-					return NULL;
-				}
-			}
-
-			delete io;
-		}
-	}
-
-	return NULL;
-}
-
 FIMULTIBITMAP * DLL_CALLCONV
 FreeImage_OpenMultiBitmap(FREE_IMAGE_FORMAT fif, const char *filename, BOOL create_new, BOOL read_only, BOOL keep_cache_in_memory, int flags) {
-	// sanity check on the parameters
 
-	if (create_new)
-		read_only = FALSE;
+	FILE *handle = NULL;
+	try {
+		// sanity check on the parameters
 
-	// retrieve the plugin list to find the node belonging to this plugin
+		if (create_new) {
+			read_only = FALSE;
+		}
 
-	PluginList *list = FreeImage_GetPluginList();
+		// retrieve the plugin list to find the node belonging to this plugin
 
-	if (list) {
-		PluginNode *node = list->FindNodeFromFIF(fif);
+		PluginList *list = FreeImage_GetPluginList();
 
-		if (node) {
-			FreeImageIO *io = new FreeImageIO;
+		if (list) {
+			PluginNode *node = list->FindNodeFromFIF(fif);
 
-			if (io) {
-				SetDefaultIO(io);
+			if (node) {
+				std::auto_ptr<FreeImageIO> io (new FreeImageIO);
 
-				BOOL cont = TRUE;
-
-				FILE *handle = NULL;
+				SetDefaultIO(io.get());
 
 				if (!create_new) {
 					handle = fopen(filename, "rb");
-
 					if (handle == NULL) {
-						cont = FALSE;
-					}
-				}
-
-				if (cont) {
-					FIMULTIBITMAP *bitmap = new FIMULTIBITMAP;
-
-					if (bitmap) {
-						MULTIBITMAPHEADER *header = new MULTIBITMAPHEADER;
-
-						header->m_filename = new char[strlen(filename) + 1];
-						strcpy(header->m_filename, filename);
-						header->node = node;
-						header->fif = fif;
-						header->io = io;
-						header->handle = handle;						
-						header->changed = FALSE;						
-						header->read_only = read_only;
-						header->m_cachefile = NULL;
-						header->cache_fif = fif;
-						header->load_flags = flags;
-
-						if (header) {
-							// store the MULTIBITMAPHEADER in the surrounding FIMULTIBITMAP structure
-
-							bitmap->data = header;
-
-							// cache the page count
-
-							header->page_count = FreeImage_InternalGetPageCount(bitmap);
-
-							// allocate a continueus block to describe the bitmap
-
-							if (!create_new)
-								header->m_blocks.push_back((BlockTypeS *)new BlockContinueus(0, header->page_count - 1));
-
-							// set up the cache
-
-							if (!read_only) {
-								char cache_name[256];
-								ReplaceExtension(cache_name, filename, "ficache");
-
-								CacheFile *cache_file = new CacheFile(cache_name, keep_cache_in_memory);
-
-								if (cache_file->open()) {
-									header->m_cachefile = cache_file;
-
-									// return the multibitmap
-
-									return bitmap;
-								}
-
-								delete cache_file;
-								delete header;
-							}
-
-							return bitmap;
-						}
-						
 						return NULL;
 					}
 				}
+
+				std::auto_ptr<FIMULTIBITMAP> bitmap (new FIMULTIBITMAP);
+				std::auto_ptr<MULTIBITMAPHEADER> header (new MULTIBITMAPHEADER);
+				header->m_filename = new char[strlen(filename) + 1];
+				strcpy(header->m_filename, filename);
+				header->node = node;
+				header->fif = fif;
+				header->io = io.get ();
+				header->handle = handle;						
+				header->changed = FALSE;						
+				header->read_only = read_only;
+				header->m_cachefile = NULL;
+				header->cache_fif = fif;
+				header->load_flags = flags;
+
+				// store the MULTIBITMAPHEADER in the surrounding FIMULTIBITMAP structure
+
+				bitmap->data = header.get();
+
+				// cache the page count
+
+				header->page_count = FreeImage_InternalGetPageCount(bitmap.get());
+
+				// allocate a continueus block to describe the bitmap
+
+				if (!create_new) {
+					header->m_blocks.push_back((BlockTypeS *)new BlockContinueus(0, header->page_count - 1));
+				}
+
+				// set up the cache
+
+				if (!read_only) {
+					std::string cache_name;
+					ReplaceExtension(cache_name, filename, "ficache");
+
+					std::auto_ptr<CacheFile> cache_file (new CacheFile(cache_name, keep_cache_in_memory));
+
+					if (cache_file->open()) {
+						// we can use release() as std::bad_alloc won't be thrown from here on
+						header->m_cachefile = cache_file.release();
+					} else {
+						// an error occured ...
+						fclose(handle);
+						return NULL;
+					}
+				}
+				// return the multibitmap
+				// std::bad_alloc won't be thrown from here on
+				header.release(); // now owned by bitmap
+				io.release();	  // now owned by bitmap
+				return bitmap.release(); // now owned by caller
 			}
-
-			delete io;
 		}
+	} catch (std::bad_alloc &) {
+		/** @todo report error */
 	}
-
+	if (handle)
+		fclose(handle);
 	return NULL;
 }
 
 FIMULTIBITMAP * DLL_CALLCONV
 FreeImage_OpenMultiBitmapFromHandle(FREE_IMAGE_FORMAT fif, FreeImageIO *io, fi_handle handle, int flags) {
-	
-	if (io && handle) {
+	try {
+		BOOL read_only = FALSE;	// modifications (if any) will be stored into the memory cache
+
+		if (io && handle) {
 		
-		// retrieve the plugin list to find the node belonging to this plugin
-		PluginList *list = FreeImage_GetPluginList();
+			// retrieve the plugin list to find the node belonging to this plugin
+			PluginList *list = FreeImage_GetPluginList();
 		
-		if (list) {
-			PluginNode *node = list->FindNodeFromFIF(fif);
+			if (list) {
+				PluginNode *node = list->FindNodeFromFIF(fif);
 			
-			if (node) {
-				FIMULTIBITMAP *bitmap = new FIMULTIBITMAP;
-				
-				if (bitmap) {
-					MULTIBITMAPHEADER *header = new MULTIBITMAPHEADER;
-					
-					if (header) {
-						header->io = new FreeImageIO;
+				if (node) {
+					std::auto_ptr<FIMULTIBITMAP> bitmap (new FIMULTIBITMAP);
+					std::auto_ptr<MULTIBITMAPHEADER> header (new MULTIBITMAPHEADER);
+					std::auto_ptr<FreeImageIO> tmp_io (new FreeImageIO (*io));
+					header->io = tmp_io.get();
+					header->m_filename = NULL;
+					header->node = node;
+					header->fif = fif;
+					header->handle = handle;						
+					header->changed = FALSE;						
+					header->read_only = read_only;	
+					header->m_cachefile = NULL;
+					header->cache_fif = fif;
+					header->load_flags = flags;
+							
+					// store the MULTIBITMAPHEADER in the surrounding FIMULTIBITMAP structure
+
+					bitmap->data = header.get();
+
+					// cache the page count
+
+					header->page_count = FreeImage_InternalGetPageCount(bitmap.get());
+
+					// allocate a continueus block to describe the bitmap
+
+					header->m_blocks.push_back((BlockTypeS *)new BlockContinueus(0, header->page_count - 1));
+
+					if (!read_only) {
+						// set up the cache
+						std::auto_ptr<CacheFile> cache_file (new CacheFile("", TRUE));
 						
-						if (header->io) {
-							
-							header->m_filename = NULL;
-							header->node = node;
-							header->fif = fif;
-							*header->io = *io;
-							header->handle = handle;						
-							header->changed = FALSE;						
-							header->read_only = TRUE;
-							header->m_cachefile = NULL;
-							header->cache_fif = fif;
-							header->load_flags = flags;
-							
-							// store the MULTIBITMAPHEADER in the surrounding FIMULTIBITMAP structure
-
-							bitmap->data = header;
-
-							// cache the page count
-
-							header->page_count = FreeImage_InternalGetPageCount(bitmap);
-
-							// allocate a continueus block to describe the bitmap
-
-							header->m_blocks.push_back((BlockTypeS *)new BlockContinueus(0, header->page_count - 1));
-
-							return bitmap;
+						if (cache_file->open()) {
+							header->m_cachefile = cache_file.release();
 						}
-						delete header;
 					}
-					delete bitmap;
+					tmp_io.release();
+					header.release();
+					return bitmap.release();
 				}
 			}
 		}
+	} catch (std::bad_alloc &) {
+		/** @todo report error */
 	}
-
 	return NULL;
 }
 
+BOOL DLL_CALLCONV
+FreeImage_SaveMultiBitmapToHandle(FREE_IMAGE_FORMAT fif, FIMULTIBITMAP *bitmap, FreeImageIO *io, fi_handle handle, int flags) {
+	if(!bitmap || !bitmap->data || !io || !handle) {
+		return FALSE;
+	}
+
+	BOOL success = TRUE;
+
+	// retrieve the plugin list to find the node belonging to this plugin
+	PluginList *list = FreeImage_GetPluginList();
+	
+	if (list) {
+		PluginNode *node = list->FindNodeFromFIF(fif);
+
+		if(node) {
+			MULTIBITMAPHEADER *header = FreeImage_GetMultiBitmapHeader(bitmap);
+			
+			// dst data
+			void *data = FreeImage_Open(node, io, handle, FALSE);
+			// src data
+			void *data_read = NULL;
+			
+			if(header->handle) {
+				// open src
+				header->io->seek_proc(header->handle, 0, SEEK_SET);
+				data_read = FreeImage_Open(header->node, header->io, header->handle, TRUE);
+			}
+			
+			// write all the pages to the file using handle and io
+			
+			int count = 0;
+			
+			for (BlockListIterator i = header->m_blocks.begin(); i != header->m_blocks.end(); i++) {
+				if (success) {
+					switch((*i)->m_type) {
+						case BLOCK_CONTINUEUS:
+						{
+							BlockContinueus *block = (BlockContinueus *)(*i);
+							
+							for (int j = block->m_start; j <= block->m_end; j++) {
+
+								// load the original source data
+								FIBITMAP *dib = header->node->m_plugin->load_proc(header->io, header->handle, j, header->load_flags, data_read);
+								
+								// save the data
+								success = node->m_plugin->save_proc(io, dib, handle, count, flags, data);
+								count++;
+								
+								FreeImage_Unload(dib);
+							}
+							
+							break;
+						}
+						
+						case BLOCK_REFERENCE:
+						{
+							BlockReference *ref = (BlockReference *)(*i);
+							
+							// read the compressed data
+							
+							BYTE *compressed_data = (BYTE*)malloc(ref->m_size * sizeof(BYTE));
+							
+							header->m_cachefile->readFile((BYTE *)compressed_data, ref->m_reference, ref->m_size);
+							
+							// uncompress the data
+							
+							FIMEMORY *hmem = FreeImage_OpenMemory(compressed_data, ref->m_size);
+							FIBITMAP *dib = FreeImage_LoadFromMemory(header->cache_fif, hmem, 0);
+							FreeImage_CloseMemory(hmem);
+							
+							// get rid of the buffer
+							free(compressed_data);
+							
+							// save the data
+							
+							success = node->m_plugin->save_proc(io, dib, handle, count, flags, data);
+							count++;
+							
+							// unload the dib
+
+							FreeImage_Unload(dib);
+
+							break;
+						}
+					}
+				} else {
+					break;
+				}
+			}
+			
+			// close the files
+			
+			FreeImage_Close(header->node, header->io, header->handle, data_read);
+
+			FreeImage_Close(node, io, handle, data); 
+			
+			return success;
+		}
+	}
+
+	return FALSE;
+}
+
+
 BOOL DLL_CALLCONV
 FreeImage_CloseMultiBitmap(FIMULTIBITMAP *bitmap, int flags) {
 	if (bitmap) {
 		BOOL success = TRUE;
-
+		
 		if (bitmap->data) {
 			MULTIBITMAPHEADER *header = FreeImage_GetMultiBitmapHeader(bitmap);			
+			
+			// saves changes only of images loaded directly from a file
+			if (header->changed && header->m_filename) {
+				try {
+					// open a temp file
 
-			if (header->changed) {
-				// open a temp file
+					std::string spool_name;
 
-				char spool_name[256];
+					ReplaceExtension(spool_name, header->m_filename, "fispool");
 
-				ReplaceExtension(spool_name, header->m_filename, "fispool");
+					// open the spool file and the source file
+        
+					FILE *f = fopen(spool_name.c_str(), "w+b");
+				
+					// saves changes
+					if (f == NULL) {
+						FreeImage_OutputMessageProc(header->fif, "Failed to open %s, %s", spool_name.c_str(), strerror(errno));
+						success = FALSE;
+					} else {
+						success = FreeImage_SaveMultiBitmapToHandle(header->fif, bitmap, header->io, (fi_handle)f, flags);
 
-				// open the spool file and the source file
+						// close the files
 
-				FILE *f = fopen(spool_name, "w+b");
+						if (fclose(f) != 0) {
+							success = FALSE;
+							FreeImage_OutputMessageProc(header->fif, "Failed to close %s, %s", spool_name.c_str(), strerror(errno));
+						}
+					}
+					if (header->handle) {
+						fclose((FILE *)header->handle);
+					}
+				
+					// applies changes to the destination file
 
-				void *data = FreeImage_Open(header->node, header->io, (fi_handle)f, FALSE);
-				void *data_read = NULL;
-
-				if (header->handle) {
-					header->io->seek_proc(header->handle, 0, SEEK_SET);
-
-			   		data_read = FreeImage_Open(header->node, header->io, header->handle, TRUE);
-				}
-
-				// write all the pages to the temp file
-
-				int count = 0;				
-
-				for (BlockListIterator i = header->m_blocks.begin(); i != header->m_blocks.end(); i++) {
 					if (success) {
-						switch((*i)->m_type) {
-							case BLOCK_CONTINUEUS :
-							{
-								BlockContinueus *block = (BlockContinueus *)(*i);
-
-								for (int j = block->m_start; j <= block->m_end; j++) {
-									FIBITMAP *dib = header->node->m_plugin->load_proc(header->io, header->handle, j, header->load_flags, data_read);
-
-									success = header->node->m_plugin->save_proc(header->io, dib, (fi_handle)f, count, flags, data);
-									count++;
-
-									FreeImage_Unload(dib);
-								}
-
-								break;
-							}
-
-							case BLOCK_REFERENCE :
-							{
-								BlockReference *ref = (BlockReference *)(*i);
-
-								// read the compressed data
-
-								BYTE *compressed_data = (BYTE*)malloc(ref->m_size * sizeof(BYTE));
-
-								header->m_cachefile->readFile((BYTE *)compressed_data, ref->m_reference, ref->m_size);
-
-								// uncompress the data
-
-								FIMEMORY *hmem = FreeImage_OpenMemory(compressed_data, ref->m_size);
-								FIBITMAP *dib = FreeImage_LoadFromMemory(header->cache_fif, hmem, 0);
-								FreeImage_CloseMemory(hmem);
-
-								// get rid of the buffer
-								free(compressed_data);
-
-								// save the data
-
-								success = header->node->m_plugin->save_proc(header->io, dib, (fi_handle)f, count, flags, data);
-								count++;
-
-								// unload the dib
-
-								FreeImage_Unload(dib);
-
-								break;
-							}
+						remove(header->m_filename);
+						success = (rename(spool_name.c_str(), header->m_filename) == 0) ? TRUE:FALSE;
+						if(!success) {
+							FreeImage_OutputMessageProc(header->fif, "Failed to rename %s to %s", spool_name.c_str(), header->m_filename);
 						}
 					} else {
-						break;
+						remove(spool_name.c_str());
 					}
+				} catch (std::bad_alloc &) {
+					success = FALSE;
 				}
 
-				// close the files
-
-				FreeImage_Close(header->node, header->io, (fi_handle)f, data); 
-
-				fclose(f);
-
-				if (header->handle) {
-					FreeImage_Close(header->node, header->io, header->handle, data_read);
-
-					fclose((FILE *)header->handle);
-				}	
-
-				if (success) {
-					remove(header->m_filename);
-
-					rename(spool_name, header->m_filename);
-				} else {
-					remove(spool_name);
-				}
 			} else {
 				if (header->handle && header->m_filename) {
 					fclose((FILE *)header->handle);
@@ -543,14 +548,14 @@ FreeImage_CloseMultiBitmap(FIMULTIBITMAP *bitmap, int flags) {
 
 			// clear the blocks list
 
-			for (BlockListIterator i = header->m_blocks.begin(); i != header->m_blocks.end(); ++i)
+			for (BlockListIterator i = header->m_blocks.begin(); i != header->m_blocks.end(); ++i) {
 				delete *i;
+			}
 
 			// flush and dispose the cache
 
 			if (header->m_cachefile) {
 				header->m_cachefile->close();
-
 				delete header->m_cachefile;
 			}
 
@@ -568,8 +573,9 @@ FreeImage_CloseMultiBitmap(FIMULTIBITMAP *bitmap, int flags) {
 
 			// delete the filename
 
-			if(header->m_filename)
+			if(header->m_filename) {
 				delete[] header->m_filename;
+			}
 
 			// delete the FIMULTIBITMAPHEADER
 
@@ -708,11 +714,13 @@ FreeImage_DeletePage(FIMULTIBITMAP *bitmap, int page) {
 				if (i != header->m_blocks.end()) {
 					switch((*i)->m_type) {
 						case BLOCK_CONTINUEUS :
+							delete *i;
 							header->m_blocks.erase(i);
 							break;
 
 						case BLOCK_REFERENCE :
 							header->m_cachefile->deleteFile(((BlockReference *)(*i))->m_reference);
+							delete *i;
 							header->m_blocks.erase(i);
 							break;
 					}
@@ -890,3 +898,88 @@ FreeImage_GetLockedPageNumbers(FIMULTIBITMAP *bitmap, int *pages, int *count) {
 
 	return FALSE;
 }
+
+// =====================================================================
+// Memory IO Multipage functions
+// =====================================================================
+
+FIMULTIBITMAP * DLL_CALLCONV
+FreeImage_LoadMultiBitmapFromMemory(FREE_IMAGE_FORMAT fif, FIMEMORY *stream, int flags) {
+	BOOL read_only = FALSE;	// modifications (if any) will be stored into the memory cache
+
+	// retrieve the plugin list to find the node belonging to this plugin
+
+	PluginList *list = FreeImage_GetPluginList();
+
+	if (list) {
+		PluginNode *node = list->FindNodeFromFIF(fif);
+
+		if (node) {
+			FreeImageIO *io = new(std::nothrow) FreeImageIO;
+
+			if (io) {
+				SetMemoryIO(io);
+
+				FIMULTIBITMAP *bitmap = new(std::nothrow) FIMULTIBITMAP;
+
+				if (bitmap) {
+					MULTIBITMAPHEADER *header = new(std::nothrow) MULTIBITMAPHEADER;
+
+					if (header) {
+						header->m_filename = NULL;
+						header->node = node;
+						header->fif = fif;
+						header->io = io;
+						header->handle = (fi_handle)stream;						
+						header->changed = FALSE;						
+						header->read_only = read_only;
+						header->m_cachefile = NULL;
+						header->cache_fif = fif;
+						header->load_flags = flags;
+
+						// store the MULTIBITMAPHEADER in the surrounding FIMULTIBITMAP structure
+
+						bitmap->data = header;
+
+						// cache the page count
+
+						header->page_count = FreeImage_InternalGetPageCount(bitmap);
+
+						// allocate a continueus block to describe the bitmap
+
+						header->m_blocks.push_back((BlockTypeS *)new BlockContinueus(0, header->page_count - 1));
+						
+						if (!read_only) {
+							// set up the cache
+							CacheFile *cache_file = new(std::nothrow) CacheFile("", TRUE);
+							
+							if (cache_file && cache_file->open()) {
+								header->m_cachefile = cache_file;
+							}
+						}
+
+						return bitmap;
+					}
+					
+					delete bitmap;
+				}
+				
+				delete io;
+			}
+		}
+	}
+
+	return NULL;
+}
+
+BOOL DLL_CALLCONV
+FreeImage_SaveMultiBitmapToMemory(FREE_IMAGE_FORMAT fif, FIMULTIBITMAP *bitmap, FIMEMORY *stream, int flags) {
+	if (stream && stream->data) {
+		FreeImageIO io;
+		SetMemoryIO(&io);
+
+		return FreeImage_SaveMultiBitmapToHandle(fif, bitmap, &io, (fi_handle)stream, flags);
+	}
+
+	return FALSE;
+}
