@@ -90,9 +90,10 @@ PluginList::AddNode(FI_InitProc init_proc, void *instance, const char *format, c
 
 		memset(plugin, 0, sizeof(Plugin));
 
-		// fill-in the plugin structure
+		// fill-in the plugin structure 
+		// note we have memset to 0, so all unset pointers should be NULL)
 
-		init_proc(plugin, m_plugin_map.size());
+		init_proc(plugin, (int)m_plugin_map.size());
 
 		// get the format string (two possible ways)
 
@@ -107,17 +108,16 @@ PluginList::AddNode(FI_InitProc init_proc, void *instance, const char *format, c
 
 		if (the_format != NULL) {
 			if (FindNodeFromFormat(the_format) == NULL) {
-				node->m_id = m_plugin_map.size();
+				node->m_id = (int)m_plugin_map.size();
 				node->m_instance = instance;
 				node->m_plugin = plugin;
 				node->m_format = format;
 				node->m_description = description;
 				node->m_extension = extension;
 				node->m_regexpr = regexpr;
-				node->m_next = NULL;
 				node->m_enabled = TRUE;
 
-				m_plugin_map[m_plugin_map.size()] = node;
+				m_plugin_map[(const int)m_plugin_map.size()] = node;
 
 				return (FREE_IMAGE_FORMAT)node->m_id;
 			}
@@ -176,7 +176,7 @@ PluginList::FindNodeFromFIF(int node_id) {
 
 int
 PluginList::Size() const {
-	return m_plugin_map.size();
+	return (int)m_plugin_map.size();
 }
 
 BOOL
@@ -214,7 +214,7 @@ FreeImage_Initialise(BOOL load_local_plugins_only) {
 
 		// internal plugin initialization
 
-		s_plugins = new PluginList;
+		s_plugins = new(std::nothrow) PluginList;
 
 		if (s_plugins) {
 			/* NOTE : 
@@ -253,6 +253,9 @@ FreeImage_Initialise(BOOL load_local_plugins_only) {
 			s_plugins->AddNode(InitEXR);
 			s_plugins->AddNode(InitJ2K);
 			s_plugins->AddNode(InitJP2);
+			s_plugins->AddNode(InitPFM);
+			s_plugins->AddNode(InitPICT);
+			s_plugins->AddNode(InitRAW);
 			
 			// external plugin initialization
 
@@ -286,7 +289,7 @@ FreeImage_Initialise(BOOL load_local_plugins_only) {
 					strcpy(buffer, s_search_list[count]);
 					strcat(buffer, "*.fip");
 
-					if ((find_handle= _findfirst(buffer, &find_data)) != -1L) {
+					if ((find_handle = (long)_findfirst(buffer, &find_data)) != -1L) {
 						do {
 							strcpy(buffer, s_search_list[count]);
 							strncat(buffer, find_data.name, MAX_PATH + 200);
@@ -296,10 +299,11 @@ FreeImage_Initialise(BOOL load_local_plugins_only) {
 							if (instance != NULL) {
 								FARPROC proc_address = GetProcAddress(instance, "_Init@8");
 
-								if (proc_address != NULL)
+								if (proc_address != NULL) {
 									s_plugins->AddNode((FI_InitProc)proc_address, (void *)instance);
-								else
+								} else {
 									FreeLibrary(instance);
+								}
 							}
 						} while (_findnext(find_handle, &find_data) != -1L);
 
@@ -311,8 +315,9 @@ FreeImage_Initialise(BOOL load_local_plugins_only) {
 
 				// restore the current directory
 
-				if (bOk)
+				if (bOk) {
 					SetCurrentDirectory(current_dir);
+				}
 			}
 #endif // _WIN32
 		}
@@ -334,17 +339,19 @@ FreeImage_DeInitialise() {
 
 void * DLL_CALLCONV
 FreeImage_Open(PluginNode *node, FreeImageIO *io, fi_handle handle, BOOL open_for_reading) {
-	if (node->m_plugin->open_proc != NULL)
+	if (node->m_plugin->open_proc != NULL) {
        return node->m_plugin->open_proc(io, handle, open_for_reading);
+	}
 
 	return NULL;
-};
+}
 
 void DLL_CALLCONV
 FreeImage_Close(PluginNode *node, FreeImageIO *io, fi_handle handle, void *data) {
-	if (node->m_plugin->close_proc != NULL)
+	if (node->m_plugin->close_proc != NULL) {
 		node->m_plugin->close_proc(io, handle, data);
-};
+	}
+}
 
 // =====================================================================
 // Plugin System Load/Save Functions
@@ -388,6 +395,8 @@ FreeImage_Load(FREE_IMAGE_FORMAT fif, const char *filename, int flags) {
 		fclose(handle);
 
 		return bitmap;
+	} else {
+		FreeImage_OutputMessageProc((int)fif, "FreeImage_Load: failed to open file %s", filename);
 	}
 
 	return NULL;
@@ -406,6 +415,8 @@ FreeImage_LoadU(FREE_IMAGE_FORMAT fif, const wchar_t *filename, int flags) {
 		fclose(handle);
 
 		return bitmap;
+	} else {
+		FreeImage_OutputMessageProc((int)fif, "FreeImage_LoadU: failed to open input file");
 	}
 #endif
 	return NULL;
@@ -413,6 +424,12 @@ FreeImage_LoadU(FREE_IMAGE_FORMAT fif, const wchar_t *filename, int flags) {
 
 BOOL DLL_CALLCONV
 FreeImage_SaveToHandle(FREE_IMAGE_FORMAT fif, FIBITMAP *dib, FreeImageIO *io, fi_handle handle, int flags) {
+	// cannot save "header only" formats
+	if(FreeImage_HasPixels(dib) == FALSE) {
+		FreeImage_OutputMessageProc((int)fif, "FreeImage_SaveToHandle: cannot save \"header only\" formats");
+		return FALSE;
+	}
+
 	if ((fif >= 0) && (fif < FreeImage_GetFIFCount())) {
 		PluginNode *node = s_plugins->FindNodeFromFIF(fif);
 		
@@ -450,6 +467,8 @@ FreeImage_Save(FREE_IMAGE_FORMAT fif, FIBITMAP *dib, const char *filename, int f
 		fclose(handle);
 
 		return success;
+	} else {
+		FreeImage_OutputMessageProc((int)fif, "FreeImage_Save: failed to open file %s", filename);
 	}
 
 	return FALSE;
@@ -468,6 +487,8 @@ FreeImage_SaveU(FREE_IMAGE_FORMAT fif, FIBITMAP *dib, const wchar_t *filename, i
 		fclose(handle);
 
 		return success;
+	} else {
+		FreeImage_OutputMessageProc((int)fif, "FreeImage_SaveU: failed to open output file");
 	}
 #endif
 	return FALSE;
@@ -679,6 +700,19 @@ FreeImage_FIFSupportsICCProfiles(FREE_IMAGE_FORMAT fif) {
 	return FALSE;
 }
 
+BOOL DLL_CALLCONV
+FreeImage_FIFSupportsNoPixels(FREE_IMAGE_FORMAT fif) {
+	if (s_plugins != NULL) {
+		PluginNode *node = s_plugins->FindNodeFromFIF(fif);
+
+		return (node != NULL) ? 
+			(node->m_plugin->supports_no_pixels_proc != NULL) ? 
+				node->m_plugin->supports_no_pixels_proc() : FALSE : FALSE;
+	}
+
+	return FALSE;
+}
+
 FREE_IMAGE_FORMAT DLL_CALLCONV
 FreeImage_GetFIFFromFilename(const char *filename) {
 	if (filename != NULL) {
