@@ -3,6 +3,7 @@
 //
 // Design and implementation by 
 // - Herv Drolon (drolon@infonie.fr)
+// - Mihail Naydenov (mnaydenov@users.sourceforge.net)
 //
 // This file is part of FreeImage 3
 //
@@ -29,6 +30,7 @@
 #include "../OpenEXR/IlmImf/ImfChannelList.h"
 #include "../OpenEXR/IlmImf/ImfRgba.h"
 #include "../OpenEXR/IlmImf/ImfArray.h"
+#include "../OpenEXR/IlmImf/ImfPreviewImage.h"
 #include "../OpenEXR/Half/half.h"
 
 
@@ -164,6 +166,11 @@ SupportsExportType(FREE_IMAGE_TYPE type) {
 	);
 }
 
+static BOOL DLL_CALLCONV
+SupportsNoPixels() {
+	return TRUE;
+}
+
 // --------------------------------------------------------------------------
 
 static FIBITMAP * DLL_CALLCONV
@@ -171,208 +178,321 @@ Load(FreeImageIO *io, fi_handle handle, int page, int flags, void *data) {
 	bool bUseRgbaInterface = false;
 	FIBITMAP *dib = NULL;	
 
-	if(handle) {
-		try {
-			// save the stream starting point
-			long stream_start = io->tell_proc(handle);
+	if(!handle) {
+		return NULL;
+	}
 
-			// wrap the FreeImage IO stream
-			C_IStream istream(io, handle);
+	try {
+		BOOL header_only = (flags & FIF_LOAD_NOPIXELS) == FIF_LOAD_NOPIXELS;
 
-			// open the file
-			Imf::InputFile file(istream);
+		// save the stream starting point
+		long stream_start = io->tell_proc(handle);
 
-			// get file info			
-			const Imath::Box2i &dataWindow = file.header().dataWindow();
-			int width  = dataWindow.max.x - dataWindow.min.x + 1;
-			int height = dataWindow.max.y - dataWindow.min.y + 1;
+		// wrap the FreeImage IO stream
+		C_IStream istream(io, handle);
 
-			//const Imf::Compression &compression = file.header().compression();
+		// open the file
+		Imf::InputFile file(istream);
 
-			if((dataWindow.min.x > 0) || (dataWindow.min.x < 0) || (dataWindow.min.y > 0) || (dataWindow.min.y < 0)) {
-				THROW (Iex::InputExc, "Invalid data window " << 
-					"[" << dataWindow.min.x << "," << dataWindow.min.y << "," << dataWindow.max.x << "," << dataWindow.max.y << "]");
-			}
-            			
-			const Imf::ChannelList &channels = file.header().channels();
+		// get file info			
+		const Imath::Box2i &dataWindow = file.header().dataWindow();
+		int width  = dataWindow.max.x - dataWindow.min.x + 1;
+		int height = dataWindow.max.y - dataWindow.min.y + 1;
 
-			// check the number of components and check for a coherent format
+		//const Imf::Compression &compression = file.header().compression();
 
-			std::string exr_color_model;
-			Imf::PixelType pixel_type = Imf::HALF;
-			FREE_IMAGE_TYPE image_type = FIT_UNKNOWN;
-			int components = 0;
-			bool bMixedComponents = false;
+		const Imf::ChannelList &channels = file.header().channels();
 
-			for (Imf::ChannelList::ConstIterator i = channels.begin(); i != channels.end(); ++i) {
-				components++;
-				if(components == 1) {
-					exr_color_model += i.name();
-					pixel_type = i.channel().type;
-				} else {
-					exr_color_model += "/";
-					exr_color_model += i.name();
-					if (i.channel().type != pixel_type) {
-						bMixedComponents = true;
-					}
+		// check the number of components and check for a coherent format
+
+		std::string exr_color_model;
+		Imf::PixelType pixel_type = Imf::HALF;
+		FREE_IMAGE_TYPE image_type = FIT_UNKNOWN;
+		int components = 0;
+		bool bMixedComponents = false;
+
+		for (Imf::ChannelList::ConstIterator i = channels.begin(); i != channels.end(); ++i) {
+			components++;
+			if(components == 1) {
+				exr_color_model += i.name();
+				pixel_type = i.channel().type;
+			} else {
+				exr_color_model += "/";
+				exr_color_model += i.name();
+				if (i.channel().type != pixel_type) {
+					bMixedComponents = true;
 				}
 			}
-			if(bMixedComponents) {
-				bool bHandled = false;
-				// we may have a RGBZ or RGBAZ image ... 
-				if(components > 4) {
-					if(channels.findChannel("R") && channels.findChannel("G") && channels.findChannel("B") && channels.findChannel("A")) {
+		}
+
+		if(bMixedComponents) {
+			bool bHandled = false;
+			// we may have a RGBZ or RGBAZ image ... 
+			if(components > 4) {
+				if(channels.findChannel("R") && channels.findChannel("G") && channels.findChannel("B") && channels.findChannel("A")) {
+					std::string msg = "Warning: converting color model " + exr_color_model + " to RGBA color model";
+					FreeImage_OutputMessageProc(s_format_id, msg.c_str());
+					bHandled = true;
+				}
+			}
+			else if(components > 3) {
+				if(channels.findChannel("R") && channels.findChannel("G") && channels.findChannel("B")) {
+					std::string msg = "Warning: converting color model " + exr_color_model + " to RGB color model";
+					FreeImage_OutputMessageProc(s_format_id, msg.c_str());
+					bHandled = true;
+				}
+			}
+			if(!bHandled) {
+				THROW (Iex::InputExc, "Unable to handle mixed component types (color model = " << exr_color_model << ")");
+			} 
+		}
+
+		switch(pixel_type) {
+			case Imf::UINT:
+				THROW (Iex::InputExc, "Unsupported format: UINT");
+				break;
+			case Imf::HALF:
+			case Imf::FLOAT:
+			default:
+				break;
+		}
+
+		// check for supported image color models
+		// --------------------------------------------------------------
+
+		if((components == 1) || (components == 2)) {				
+			// if the image is gray-alpha (YA), ignore the alpha channel
+			if((components == 1) && channels.findChannel("Y")) {
+				image_type = FIT_FLOAT;
+				components = 1;
+			} else {
+				std::string msg = "Warning: loading color model " + exr_color_model + " as Y color model";
+				FreeImage_OutputMessageProc(s_format_id, msg.c_str());
+				image_type = FIT_FLOAT;
+				// ignore the other channel
+				components = 1;
+			}
+		} else if(components == 3) {
+			if(channels.findChannel("R") && channels.findChannel("G") && channels.findChannel("B")) {
+				image_type = FIT_RGBF;
+			}
+			else if(channels.findChannel("BY") && channels.findChannel("RY") && channels.findChannel("Y")) {
+				image_type = FIT_RGBF;
+				bUseRgbaInterface = true;
+			}
+		} else if(components >= 4) {
+			if(channels.findChannel("R") && channels.findChannel("G") && channels.findChannel("B")) {
+				if(channels.findChannel("A")) {
+					if(components > 4) {
 						std::string msg = "Warning: converting color model " + exr_color_model + " to RGBA color model";
 						FreeImage_OutputMessageProc(s_format_id, msg.c_str());
-						bHandled = true;
 					}
-				}
-				else if(components > 3) {
-					if(channels.findChannel("R") && channels.findChannel("G") && channels.findChannel("B")) {
-						std::string msg = "Warning: converting color model " + exr_color_model + " to RGB color model";
-						FreeImage_OutputMessageProc(s_format_id, msg.c_str());
-						bHandled = true;
-					}
-				}
-				if(!bHandled) {
-					THROW (Iex::InputExc, "Unable to handle mixed component types (color model = " << exr_color_model << ")");
-				} 
-			}
-			switch(pixel_type) {
-				case Imf::UINT:
-					THROW (Iex::InputExc, "Unsupported format: UINT");
-					break;
-				case Imf::HALF:
-				case Imf::FLOAT:
-				default:
-					break;
-			}
-			// check for supported image color models
-			if((components == 1) || (components == 2)) {				
-				// if the image is gray-alpha (YA), ignore the alpha channel
-				if(channels.findChannel("Y")) {
-					image_type = FIT_FLOAT;
-					components = 1;
-				} else if(components == 1) {
-					std::string msg = "Warning: loading color model " + exr_color_model + " as Y color model";
-					FreeImage_OutputMessageProc(s_format_id, msg.c_str());
-					image_type = FIT_FLOAT;
-				}
-			} else if(components == 3) {
-				if(channels.findChannel("R") && channels.findChannel("G") && channels.findChannel("B")) {
-					image_type = FIT_RGBF;
-				}
-				else if(channels.findChannel("BY") && channels.findChannel("RY") && channels.findChannel("Y")) {
-					image_type = FIT_RGBF;
-					bUseRgbaInterface = true;
-				}
-			} else if(components >= 4) {				
-				if(channels.findChannel("R") && channels.findChannel("G") && channels.findChannel("B") && channels.findChannel("A")) {
 					image_type = FIT_RGBAF;
 					// ignore other layers if there is more than one alpha layer
 					components = 4;
+				} else {
+					std::string msg = "Warning: converting color model " + exr_color_model + " to RGB color model";
+					FreeImage_OutputMessageProc(s_format_id, msg.c_str());
+
+					image_type = FIT_RGBF;
+					// ignore other channels
+					components = 3;					
 				}
 			}
-			if(image_type == FIT_UNKNOWN) {
-				THROW (Iex::InputExc, "Unsupported color model: " << exr_color_model);
-			}
+		}
 
-			// allocate a new dib
-			dib = FreeImage_AllocateT(image_type, width, height);
-			if(!dib) THROW (Iex::NullExc, FI_MSG_ERROR_MEMORY);
+		if(image_type == FIT_UNKNOWN) {
+			THROW (Iex::InputExc, "Unsupported color model: " << exr_color_model);
+		}
 
-			BYTE *bits = FreeImage_GetBits(dib);			// pointer to our pixel buffer
-			size_t bytespp = sizeof(float) * components;	// size of our pixel in bytes
-			unsigned pitch = FreeImage_GetPitch(dib);		// size of our yStride in bytes
+		// allocate a new dib
+		dib = FreeImage_AllocateHeaderT(header_only, image_type, width, height, 0);
+		if(!dib) THROW (Iex::NullExc, FI_MSG_ERROR_MEMORY);
 
-			Imf::PixelType pixelType = Imf::FLOAT;	// load as float data type;
+		// try to load the preview image
+		// --------------------------------------------------------------
+
+		if(file.header().hasPreviewImage()) {
+			const Imf::PreviewImage& preview = file.header().previewImage();
+			const unsigned thWidth = preview.width();
+			const unsigned thHeight = preview.height();
 			
-			if(bUseRgbaInterface) {
-				// use the RGBA interface
-
-				const int chunk_size = 16;
-
-				BYTE *scanline = (BYTE*)bits;
-
-				// re-open using the RGBA interface
-				io->seek_proc(handle, stream_start, SEEK_SET);
-				Imf::RgbaInputFile rgbaFile(istream);
-
-				// read the file in chunks
-				Imath::Box2i dw = dataWindow;
-				Imf::Array2D<Imf::Rgba> chunk(chunk_size, width);
-				while (dw.min.y <= dw.max.y) {
-					// read a chunk
-					rgbaFile.setFrameBuffer (&chunk[0][0] - dw.min.x - dw.min.y * width, 1, width);
-					rgbaFile.readPixels (dw.min.y, MIN(dw.min.y + chunk_size - 1, dw.max.y));
-					// fill the dib
-					const int y_max = ((dw.max.y - dw.min.y) <= chunk_size) ? (dw.max.y - dw.min.y) : chunk_size;
-					for(int y = 0; y < y_max; y++) {
-						FIRGBF *pixel = (FIRGBF*)scanline;
-						const Imf::Rgba *half_rgba = chunk[y];
-						for(int x = 0; x < width; x++) {
-							// convert from half to float
-							pixel[x].red = half_rgba[x].r;
-							pixel[x].green = half_rgba[x].g;
-							pixel[x].blue = half_rgba[x].b;
-						}
-						// next line
-						scanline += pitch;
+			FIBITMAP* thumbnail = FreeImage_Allocate(thWidth, thHeight, 32);
+			if(thumbnail) {
+				const Imf::PreviewRgba *src_line = preview.pixels();
+				BYTE *dst_line = FreeImage_GetScanLine(thumbnail, thHeight - 1);
+				const unsigned dstPitch = FreeImage_GetPitch(thumbnail);
+				
+				for (unsigned y = 0; y < thHeight; ++y) {
+					const Imf::PreviewRgba *src_pixel = src_line;
+					RGBQUAD* dst_pixel = (RGBQUAD*)dst_line;
+					
+					for(unsigned x = 0; x < thWidth; ++x) {
+						dst_pixel->rgbRed = src_pixel->r;
+						dst_pixel->rgbGreen = src_pixel->g;
+						dst_pixel->rgbBlue = src_pixel->b;
+						dst_pixel->rgbReserved = src_pixel->a;				
+						src_pixel++;
+						dst_pixel++;
 					}
-					// next chunk
-					dw.min.y += chunk_size;
+					src_line += thWidth;
+					dst_line -= dstPitch;
 				}
+				FreeImage_SetThumbnail(dib, thumbnail);
+				FreeImage_Unload(thumbnail);
+			}
+		}
 
-			} else {
-				// use the low level interface
+		if(header_only) {
+			// header only mode
+			return dib;
+		}
 
-				// build a frame buffer (i.e. what we want on output)
-				Imf::FrameBuffer frameBuffer;
+		// load pixels
+		// --------------------------------------------------------------
 
-				if(components == 1) {
-					frameBuffer.insert ("Y",	// name
-						Imf::Slice (pixelType,	// type
-						(char*)(bits),			// base
-						bytespp,				// xStride
-						pitch,					// yStride
-						1, 1,					// x/y sampling
-						0.0));					// fillValue
-				} else if((components == 3) || (components == 4)) {
-					const char *channel_name[4] = { "R", "G", "B", "A" };
+		BYTE *bits = FreeImage_GetBits(dib);			// pointer to our pixel buffer
+		size_t bytespp = sizeof(float) * components;	// size of our pixel in bytes
+		unsigned pitch = FreeImage_GetPitch(dib);		// size of our yStride in bytes
 
-					for(int c = 0; c < components; c++) {
-						frameBuffer.insert (
-							channel_name[c],					// name
-							Imf::Slice (pixelType,				// type
-							(char*)(bits + c * sizeof(float)),	// base
-							bytespp,							// xStride
-							pitch,								// yStride
-							1, 1,								// x/y sampling
-							0.0));								// fillValue
+		Imf::PixelType pixelType = Imf::FLOAT;	// load as float data type;
+		
+		if(bUseRgbaInterface) {
+			// use the RGBA interface
+
+			const int chunk_size = 16;
+
+			BYTE *scanline = (BYTE*)bits;
+
+			// re-open using the RGBA interface
+			io->seek_proc(handle, stream_start, SEEK_SET);
+			Imf::RgbaInputFile rgbaFile(istream);
+
+			// read the file in chunks
+			Imath::Box2i dw = dataWindow;
+			Imf::Array2D<Imf::Rgba> chunk(chunk_size, width);
+			while (dw.min.y <= dw.max.y) {
+				// read a chunk
+				rgbaFile.setFrameBuffer (&chunk[0][0] - dw.min.x - dw.min.y * width, 1, width);
+				rgbaFile.readPixels (dw.min.y, MIN(dw.min.y + chunk_size - 1, dw.max.y));
+				// fill the dib
+				const int y_max = ((dw.max.y - dw.min.y) <= chunk_size) ? (dw.max.y - dw.min.y) : chunk_size;
+				for(int y = 0; y < y_max; y++) {
+					FIRGBF *pixel = (FIRGBF*)scanline;
+					const Imf::Rgba *half_rgba = chunk[y];
+					for(int x = 0; x < width; x++) {
+						// convert from half to float
+						pixel[x].red = half_rgba[x].r;
+						pixel[x].green = half_rgba[x].g;
+						pixel[x].blue = half_rgba[x].b;
 					}
+					// next line
+					scanline += pitch;
 				}
-
-				// read the file
-				file.setFrameBuffer(frameBuffer);
-				file.readPixels(dataWindow.min.y, dataWindow.max.y);
+				// next chunk
+				dw.min.y += chunk_size;
 			}
 
-			// lastly, flip dib lines
-			FreeImage_FlipVertical(dib);
+		} else {
+			// use the low level interface
 
-		}
-		catch(Iex::BaseExc & e) {
-			if(dib != NULL) {
-				FreeImage_Unload(dib);
+			// build a frame buffer (i.e. what we want on output)
+			Imf::FrameBuffer frameBuffer;
+
+			// allow dataWindow with minimal bounds different form zero
+			size_t offset = - dataWindow.min.x * bytespp - dataWindow.min.y * pitch;
+
+			if(components == 1) {
+				frameBuffer.insert ("Y",	// name
+					Imf::Slice (pixelType,	// type
+					(char*)(bits + offset), // base
+					bytespp,				// xStride
+					pitch,					// yStride
+					1, 1,					// x/y sampling
+					0.0));					// fillValue
+			} else if((components == 3) || (components == 4)) {
+				const char *channel_name[4] = { "R", "G", "B", "A" };
+
+				for(int c = 0; c < components; c++) {
+					frameBuffer.insert (
+						channel_name[c],					// name
+						Imf::Slice (pixelType,				// type
+						(char*)(bits + c * sizeof(float) + offset), // base
+						bytespp,							// xStride
+						pitch,								// yStride
+						1, 1,								// x/y sampling
+						0.0));								// fillValue
+				}
 			}
-			FreeImage_OutputMessageProc(s_format_id, e.what());
-			return NULL;
+
+			// read the file
+			file.setFrameBuffer(frameBuffer);
+			file.readPixels(dataWindow.min.y, dataWindow.max.y);
 		}
-	}	
+
+		// lastly, flip dib lines
+		FreeImage_FlipVertical(dib);
+
+	}
+	catch(Iex::BaseExc & e) {
+		if(dib != NULL) {
+			FreeImage_Unload(dib);
+		}
+		FreeImage_OutputMessageProc(s_format_id, e.what());
+		return NULL;
+	}
 
 	return dib;
 }
 
+/**
+Set the preview image using the dib embedded thumbnail
+*/
+static BOOL
+SetPreviewImage(FIBITMAP *dib, Imf::Header& header) {
+	if(!FreeImage_GetThumbnail(dib)) {
+		return FALSE;
+	}
+	FIBITMAP* thumbnail = FreeImage_GetThumbnail(dib);
+
+	if((FreeImage_GetImageType(thumbnail) != FIT_BITMAP) || (FreeImage_GetBPP(thumbnail) != 32)) {
+		// invalid thumbnail - ignore it
+		FreeImage_OutputMessageProc(s_format_id, FI_MSG_WARNING_INVALID_THUMBNAIL);
+	} else {
+		const unsigned thWidth = FreeImage_GetWidth(thumbnail);
+		const unsigned thHeight = FreeImage_GetHeight(thumbnail);
+		
+		Imf::PreviewImage preview(thWidth, thHeight);
+
+		// copy thumbnail to 32-bit RGBA preview image
+		
+		const BYTE* src_line = FreeImage_GetScanLine(thumbnail, thHeight - 1);
+		Imf::PreviewRgba* dst_line = preview.pixels();
+		const unsigned srcPitch = FreeImage_GetPitch(thumbnail);
+		
+		for (unsigned y = 0; y < thHeight; y++) {
+			const RGBQUAD* src_pixel = (RGBQUAD*)src_line;
+			Imf::PreviewRgba* dst_pixel = dst_line;
+			
+			for(unsigned x = 0; x < thWidth; x++) {
+				dst_pixel->r = src_pixel->rgbRed;
+				dst_pixel->g = src_pixel->rgbGreen;
+				dst_pixel->b = src_pixel->rgbBlue;
+				dst_pixel->a = src_pixel->rgbReserved;
+				
+				src_pixel++;
+				dst_pixel++;
+			}
+			
+			src_line -= srcPitch;
+			dst_line += thWidth;
+		}
+		
+		header.setPreviewImage(preview);
+	}
+
+	return TRUE;
+}
 
 /**
 Save using EXR_LC compression (works only with RGB[A]F images)
@@ -491,6 +611,9 @@ Save(FreeImageIO *io, FIBITMAP *dib, fi_handle handle, int page, int flags, void
 			Imath::V2f(0,0), 1, 
 			Imf::INCREASING_Y, compress);        		
 
+		// handle thumbnail
+		SetPreviewImage(dib, header);
+		
 		// check for EXR_LC compression
 		if((flags & EXR_LC) == EXR_LC) {
 			return SaveAsEXR_LC(ostream, dib, header, width, height);
@@ -543,8 +666,9 @@ Save(FreeImageIO *io, FIBITMAP *dib, fi_handle handle, int page, int flags, void
 
 		if(pixelType == Imf::HALF) {
 			// convert from float to half
-			halfData = new half[width * height * components];
+			halfData = new(std::nothrow) half[width * height * components];
 			if(!halfData) THROW (Iex::NullExc, FI_MSG_ERROR_MEMORY);
+
 			for(int y = 0; y < height; y++) {
 				float *src_bits = (float*)FreeImage_GetScanLine(dib, height - 1 - y);
 				half *dst_bits = halfData + y * width * components;
@@ -636,4 +760,5 @@ InitEXR(Plugin *plugin, int format_id) {
 	plugin->supports_export_bpp_proc = SupportsExportDepth;
 	plugin->supports_export_type_proc = SupportsExportType;
 	plugin->supports_icc_profiles_proc = NULL;
+	plugin->supports_no_pixels_proc = SupportsNoPixels;
 }
