@@ -28,7 +28,6 @@
  * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  * POSSIBILITY OF SUCH DAMAGE.
  */
-
 #include "opj_includes.h"
 
 /** @defgroup JP2 JP2 - JPEG-2000 file format reader/writer */
@@ -44,7 +43,7 @@ Read box headers
 @param box
 @return Returns true if successful, returns false otherwise
 */
-static bool jp2_read_boxhdr(opj_common_ptr cinfo, opj_cio_t *cio, opj_jp2_box_t *box);
+static opj_bool jp2_read_boxhdr(opj_common_ptr cinfo, opj_cio_t *cio, opj_jp2_box_t *box);
 /*static void jp2_write_url(opj_cio_t *cio, char *Idx_file);*/
 /**
 Read the IHDR box - Image Header box
@@ -52,12 +51,11 @@ Read the IHDR box - Image Header box
 @param cio Input buffer stream
 @return Returns true if successful, returns false otherwise
 */
-static bool jp2_read_ihdr(opj_jp2_t *jp2, opj_cio_t *cio);
+static opj_bool jp2_read_ihdr(opj_jp2_t *jp2, opj_cio_t *cio);
 static void jp2_write_ihdr(opj_jp2_t *jp2, opj_cio_t *cio);
 static void jp2_write_bpcc(opj_jp2_t *jp2, opj_cio_t *cio);
-static bool jp2_read_bpcc(opj_jp2_t *jp2, opj_cio_t *cio);
+static opj_bool jp2_read_bpcc(opj_jp2_t *jp2, opj_cio_t *cio);
 static void jp2_write_colr(opj_jp2_t *jp2, opj_cio_t *cio);
-static bool jp2_read_colr(opj_jp2_t *jp2, opj_cio_t *cio);
 /**
 Write the FTYP box - File type box
 @param jp2 JP2 handle
@@ -70,9 +68,9 @@ Read the FTYP box - File type box
 @param cio Input buffer stream
 @return Returns true if successful, returns false otherwise
 */
-static bool jp2_read_ftyp(opj_jp2_t *jp2, opj_cio_t *cio);
-static int jp2_write_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, opj_image_t *image, char *index);
-static bool jp2_read_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, unsigned int *j2k_codestream_length, unsigned int *j2k_codestream_offset);
+static opj_bool jp2_read_ftyp(opj_jp2_t *jp2, opj_cio_t *cio);
+static int jp2_write_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, opj_image_t *image, opj_codestream_info_t *cstr_info);
+static opj_bool jp2_read_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, unsigned int *j2k_codestream_length, unsigned int *j2k_codestream_offset);
 static void jp2_write_jp(opj_cio_t *cio);
 /**
 Read the JP box - JPEG 2000 signature
@@ -80,29 +78,66 @@ Read the JP box - JPEG 2000 signature
 @param cio Input buffer stream
 @return Returns true if successful, returns false otherwise
 */
-static bool jp2_read_jp(opj_jp2_t *jp2, opj_cio_t *cio);
+static opj_bool jp2_read_jp(opj_jp2_t *jp2, opj_cio_t *cio);
 /**
 Decode the structure of a JP2 file
 @param jp2 JP2 handle
 @param cio Input buffer stream
+@param color Collector for profile, cdef and pclr data
 @return Returns true if successful, returns false otherwise
 */
-static bool jp2_read_struct(opj_jp2_t *jp2, opj_cio_t *cio);
-
+static opj_bool jp2_read_struct(opj_jp2_t *jp2, opj_cio_t *cio,
+	opj_jp2_color_t *color);
+/**
+Apply collected palette data
+@param color Collector for profile, cdef and pclr data
+@param image 
+*/
+static void jp2_apply_pclr(opj_jp2_color_t *color, opj_image_t *image);
+/**
+Collect palette data
+@param jp2 JP2 handle
+@param cio Input buffer stream
+@param box
+@param color Collector for profile, cdef and pclr data
+@return Returns true if successful, returns false otherwise
+*/
+static opj_bool jp2_read_pclr(opj_jp2_t *jp2, opj_cio_t *cio,
+    opj_jp2_box_t *box, opj_jp2_color_t *color);
+/**
+Collect component mapping data
+@param jp2 JP2 handle
+@param cio Input buffer stream
+@param box
+@param color Collector for profile, cdef and pclr data
+@return Returns true if successful, returns false otherwise
+*/
+static opj_bool jp2_read_cmap(opj_jp2_t *jp2, opj_cio_t *cio,
+    opj_jp2_box_t *box, opj_jp2_color_t *color);
+/**
+Collect colour specification data
+@param jp2 JP2 handle
+@param cio Input buffer stream
+@param box
+@param color Collector for profile, cdef and pclr data
+@return Returns true if successful, returns false otherwise
+*/
+static opj_bool jp2_read_colr(opj_jp2_t *jp2, opj_cio_t *cio,
+    opj_jp2_box_t *box, opj_jp2_color_t *color);
 /*@}*/
 
 /*@}*/
 
 /* ----------------------------------------------------------------------- */
 
-static bool jp2_read_boxhdr(opj_common_ptr cinfo, opj_cio_t *cio, opj_jp2_box_t *box) {
+static opj_bool jp2_read_boxhdr(opj_common_ptr cinfo, opj_cio_t *cio, opj_jp2_box_t *box) {
 	box->init_pos = cio_tell(cio);
 	box->length = cio_read(cio, 4);
 	box->type = cio_read(cio, 4);
 	if (box->length == 1) {
 		if (cio_read(cio, 4) != 0) {
 			opj_event_msg(cinfo, EVT_ERROR, "Cannot handle box sizes higher than 2^32\n");
-			return false;
+			return OPJ_FALSE;
 		}
 		box->length = cio_read(cio, 4);
 		if (box->length == 0) 
@@ -112,7 +147,7 @@ static bool jp2_read_boxhdr(opj_common_ptr cinfo, opj_cio_t *cio, opj_jp2_box_t
 		box->length = cio_numbytesleft(cio) + 8;
 	}
 	
-	return true;
+	return OPJ_TRUE;
 }
 
 #if 0
@@ -139,7 +174,7 @@ static void jp2_write_url(opj_cio_t *cio, char *Idx_file) {
 }
 #endif
 
-static bool jp2_read_ihdr(opj_jp2_t *jp2, opj_cio_t *cio) {
+static opj_bool jp2_read_ihdr(opj_jp2_t *jp2, opj_cio_t *cio) {
 	opj_jp2_box_t box;
 
 	opj_common_ptr cinfo = jp2->cinfo;
@@ -147,7 +182,7 @@ static bool jp2_read_ihdr(opj_jp2_t *jp2, opj_cio_t *cio) {
 	jp2_read_boxhdr(cinfo, cio, &box);
 	if (JP2_IHDR != box.type) {
 		opj_event_msg(cinfo, EVT_ERROR, "Expected IHDR Marker\n");
-		return false;
+		return OPJ_FALSE;
 	}
 
 	jp2->h = cio_read(cio, 4);			/* HEIGHT */
@@ -163,10 +198,10 @@ static bool jp2_read_ihdr(opj_jp2_t *jp2, opj_cio_t *cio) {
 
 	if (cio_tell(cio) - box.init_pos != box.length) {
 		opj_event_msg(cinfo, EVT_ERROR, "Error with IHDR Box\n");
-		return false;
+		return OPJ_FALSE;
 	}
 
-	return true;
+	return OPJ_TRUE;
 }
 
 static void jp2_write_ihdr(opj_jp2_t *jp2, opj_cio_t *cio) {
@@ -211,7 +246,7 @@ static void jp2_write_bpcc(opj_jp2_t *jp2, opj_cio_t *cio) {
 }
 
 
-static bool jp2_read_bpcc(opj_jp2_t *jp2, opj_cio_t *cio) {
+static opj_bool jp2_read_bpcc(opj_jp2_t *jp2, opj_cio_t *cio) {
 	unsigned int i;
 	opj_jp2_box_t box;
 
@@ -220,7 +255,7 @@ static bool jp2_read_bpcc(opj_jp2_t *jp2, opj_cio_t *cio) {
 	jp2_read_boxhdr(cinfo, cio, &box);
 	if (JP2_BPCC != box.type) {
 		opj_event_msg(cinfo, EVT_ERROR, "Expected BPCC Marker\n");
-		return false;
+		return OPJ_FALSE;
 	}
 
 	for (i = 0; i < jp2->numcomps; i++) {
@@ -229,10 +264,10 @@ static bool jp2_read_bpcc(opj_jp2_t *jp2, opj_cio_t *cio) {
 
 	if (cio_tell(cio) - box.init_pos != box.length) {
 		opj_event_msg(cinfo, EVT_ERROR, "Error with BPCC Box\n");
-		return false;
+		return OPJ_FALSE;
 	}
 
-	return true;
+	return OPJ_TRUE;
 }
 
 static void jp2_write_colr(opj_jp2_t *jp2, opj_cio_t *cio) {
@@ -246,11 +281,10 @@ static void jp2_write_colr(opj_jp2_t *jp2, opj_cio_t *cio) {
 	cio_write(cio, jp2->precedence, 1);	/* PRECEDENCE */
 	cio_write(cio, jp2->approx, 1);		/* APPROX */
 
-	if (jp2->meth == 1) {
-		cio_write(cio, jp2->enumcs, 4);	/* EnumCS */
-	} else {
-		cio_write(cio, 0, 1);			/* PROFILE (??) */
-	}
+	if(jp2->meth == 2)
+	 jp2->enumcs = 0;
+
+	cio_write(cio, jp2->enumcs, 4);	/* EnumCS */
 
 	box.length = cio_tell(cio) - box.init_pos;
 	cio_seek(cio, box.init_pos);
@@ -258,42 +292,459 @@ static void jp2_write_colr(opj_jp2_t *jp2, opj_cio_t *cio) {
 	cio_seek(cio, box.init_pos + box.length);
 }
 
-static bool jp2_read_colr(opj_jp2_t *jp2, opj_cio_t *cio) {
-	opj_jp2_box_t box;
+static void jp2_free_pclr(opj_jp2_color_t *color)
+{
+    opj_free(color->jp2_pclr->channel_sign);
+    opj_free(color->jp2_pclr->channel_size);
+    opj_free(color->jp2_pclr->entries);
+
+	if(color->jp2_pclr->cmap) opj_free(color->jp2_pclr->cmap);
+
+    opj_free(color->jp2_pclr); color->jp2_pclr = NULL;
+}
+
+static void free_color_data(opj_jp2_color_t *color)
+{
+	if(color->jp2_pclr)
+   {
+	jp2_free_pclr(color);
+   }
+	if(color->jp2_cdef) 
+   {
+	if(color->jp2_cdef->info) opj_free(color->jp2_cdef->info);
+	opj_free(color->jp2_cdef);
+   }
+	if(color->icc_profile_buf) opj_free(color->icc_profile_buf);
+}
+
+static void jp2_apply_pclr(opj_jp2_color_t *color, opj_image_t *image)
+{
+	opj_image_comp_t *old_comps, *new_comps;
+	unsigned char *channel_size, *channel_sign;
+	unsigned int *entries;
+	opj_jp2_cmap_comp_t *cmap;
+	int *src, *dst;
+	unsigned int j, max;
+	unsigned short i, nr_channels, cmp, pcol;
+	int k, top_k;
+
+	channel_size = color->jp2_pclr->channel_size;
+	channel_sign = color->jp2_pclr->channel_sign;
+	entries = color->jp2_pclr->entries;
+	cmap = color->jp2_pclr->cmap;
+	nr_channels = color->jp2_pclr->nr_channels;
+
+	old_comps = image->comps;
+	new_comps = (opj_image_comp_t*)
+	 opj_malloc(nr_channels * sizeof(opj_image_comp_t));
+
+	for(i = 0; i < nr_channels; ++i)
+   {
+	pcol = cmap[i].pcol; cmp = cmap[i].cmp;
+
+	new_comps[pcol] = old_comps[cmp];
+
+	if(cmap[i].mtyp == 0) /* Direct use */
+  {
+	old_comps[cmp].data = NULL; continue;
+  }
+/* Palette mapping: */
+	new_comps[pcol].data = (int*)
+	 opj_malloc(old_comps[cmp].w * old_comps[cmp].h * sizeof(int));
+	new_comps[pcol].prec = channel_size[i];
+	new_comps[pcol].sgnd = channel_sign[i];
+   }
+	top_k = color->jp2_pclr->nr_entries - 1;
+
+	for(i = 0; i < nr_channels; ++i)
+   {
+/* Direct use: */
+	if(cmap[i].mtyp == 0) continue;
+
+/* Palette mapping: */
+	cmp = cmap[i].cmp; pcol = cmap[i].pcol;
+	src = old_comps[cmp].data; 
+	dst = new_comps[pcol].data;
+	max = new_comps[pcol].w * new_comps[pcol].h;
+
+	for(j = 0; j < max; ++j)
+  {
+/* The index */
+	if((k = src[j]) < 0) k = 0; else if(k > top_k) k = top_k;
+/* The colour */
+	dst[j] = entries[k * nr_channels + pcol];
+  }
+   }
+	max = image->numcomps;
+	for(i = 0; i < max; ++i)
+   {
+	if(old_comps[i].data) opj_free(old_comps[i].data);
+   }
+	opj_free(old_comps);
+	image->comps = new_comps;
+	image->numcomps = nr_channels;
+
+	jp2_free_pclr(color);
+
+}/* apply_pclr() */
+
+static opj_bool jp2_read_pclr(opj_jp2_t *jp2, opj_cio_t *cio,
+	opj_jp2_box_t *box, opj_jp2_color_t *color)
+{
+	opj_jp2_pclr_t *jp2_pclr;
+	unsigned char *channel_size, *channel_sign;
+	unsigned int *entries;
+	unsigned short nr_entries, nr_channels;
+	unsigned short i, j;
+	unsigned char uc;
+
+/* Part 1, I.5.3.4: 'There shall be at most one Palette box inside
+ * a JP2 Header box' :
+*/
+	if(color->jp2_pclr) return OPJ_FALSE;
+
+	nr_entries = cio_read(cio, 2); /* NE */
+	nr_channels = cio_read(cio, 1);/* NPC */
+
+	entries = (unsigned int*)
+	 opj_malloc(nr_channels * nr_entries * sizeof(unsigned int));
+	channel_size = (unsigned char*)opj_malloc(nr_channels);
+	channel_sign = (unsigned char*)opj_malloc(nr_channels);
+
+	jp2_pclr = (opj_jp2_pclr_t*)opj_malloc(sizeof(opj_jp2_pclr_t));
+	jp2_pclr->channel_sign = channel_sign;
+	jp2_pclr->channel_size = channel_size;
+	jp2_pclr->entries = entries;
+	jp2_pclr->nr_entries = nr_entries;
+	jp2_pclr->nr_channels = nr_channels;
+	jp2_pclr->cmap = NULL;
+
+	color->jp2_pclr = jp2_pclr;
+
+	for(i = 0; i < nr_channels; ++i)
+   {
+	uc = cio_read(cio, 1); /* Bi */
+	channel_size[i] = (uc & 0x7f) + 1;
+	channel_sign[i] = (uc & 0x80)?1:0;
+   }
+
+	for(j = 0; j < nr_entries; ++j)
+   {
+	for(i = 0; i < nr_channels; ++i)
+  {
+/* Cji */
+	*entries++ = cio_read(cio, channel_size[i]>>3);
+  }
+   }
+
+	return OPJ_TRUE;
+}/* jp2_read_pclr() */
+
+static opj_bool jp2_read_cmap(opj_jp2_t *jp2, opj_cio_t *cio,
+	opj_jp2_box_t *box, opj_jp2_color_t *color)
+{
+	opj_jp2_cmap_comp_t *cmap;
+	unsigned short i, nr_channels;
+
+/* Need nr_channels: */
+	if(color->jp2_pclr == NULL) return OPJ_FALSE;
+
+/* Part 1, I.5.3.5: 'There shall be at most one Component Mapping box
+ * inside a JP2 Header box' :
+*/
+	if(color->jp2_pclr->cmap) return OPJ_FALSE;
+
+	nr_channels = color->jp2_pclr->nr_channels;
+	cmap = (opj_jp2_cmap_comp_t*)
+	 opj_malloc(nr_channels * sizeof(opj_jp2_cmap_comp_t));
+
+	for(i = 0; i < nr_channels; ++i)
+   {
+	cmap[i].cmp = cio_read(cio, 2);
+	cmap[i].mtyp = cio_read(cio, 1);
+	cmap[i].pcol = cio_read(cio, 1);
+
+   }
+	color->jp2_pclr->cmap = cmap;
+
+	return OPJ_TRUE;
+}/* jp2_read_cmap() */
+
+static void jp2_apply_cdef(opj_image_t *image, opj_jp2_color_t *color)
+{
+	opj_jp2_cdef_info_t *info;
+	int color_space;
+	unsigned short i, n, cn, typ, asoc, acn;
+
+	color_space = image->color_space;
+	info = color->jp2_cdef->info;
+	n = color->jp2_cdef->n;
+
+	for(i = 0; i < n; ++i)
+   {
+/* WATCH: acn = asoc - 1 ! */
+	if((asoc = info[i].asoc) == 0) continue;
+
+	cn = info[i].cn; typ = info[i].typ; acn = asoc - 1;
+
+	if(cn != acn)
+  {
+	opj_image_comp_t saved;
+
+	memcpy(&saved, &image->comps[cn], sizeof(opj_image_comp_t));
+	memcpy(&image->comps[cn], &image->comps[acn], sizeof(opj_image_comp_t));
+	memcpy(&image->comps[acn], &saved, sizeof(opj_image_comp_t));
+
+	info[i].asoc = cn + 1;
+	info[acn].asoc = info[acn].cn + 1;
+  }
+   }
+	if(color->jp2_cdef->info) opj_free(color->jp2_cdef->info);
+
+	opj_free(color->jp2_cdef); color->jp2_cdef = NULL;
+
+}/* jp2_apply_cdef() */
+
+static opj_bool jp2_read_cdef(opj_jp2_t *jp2, opj_cio_t *cio,
+	opj_jp2_box_t *box, opj_jp2_color_t *color)
+{
+	opj_jp2_cdef_info_t *info;
+	unsigned short i, n;
+
+/* Part 1, I.5.3.6: 'The shall be at most one Channel Definition box
+ * inside a JP2 Header box.' 
+*/
+	if(color->jp2_cdef) return OPJ_FALSE;
+
+	if((n = cio_read(cio, 2)) == 0) return OPJ_FALSE; /* szukw000: FIXME */
+
+	info = (opj_jp2_cdef_info_t*)
+	 opj_malloc(n * sizeof(opj_jp2_cdef_info_t));
+
+	color->jp2_cdef = (opj_jp2_cdef_t*)opj_malloc(sizeof(opj_jp2_cdef_t));
+	color->jp2_cdef->info = info;
+	color->jp2_cdef->n = n;
+
+	for(i = 0; i < n; ++i)
+   {
+	info[i].cn = cio_read(cio, 2);
+	info[i].typ = cio_read(cio, 2);
+	info[i].asoc = cio_read(cio, 2);
+
+   }
+	return OPJ_TRUE;
+}/* jp2_read_cdef() */
+
+static opj_bool jp2_read_colr(opj_jp2_t *jp2, opj_cio_t *cio,
+	opj_jp2_box_t *box, opj_jp2_color_t *color) 
+{
 	int skip_len;
+    opj_common_ptr cinfo;
 
-	opj_common_ptr cinfo = jp2->cinfo;
+/* Part 1, I.5.3.3 : 'A conforming JP2 reader shall ignore all Colour
+ * Specification boxes after the first.' 
+*/
+	if(color->jp2_has_colr) return OPJ_FALSE;
 
-	jp2_read_boxhdr(cinfo, cio, &box);
-	do {
-		if (JP2_COLR != box.type) {
-			cio_skip(cio, box.length - 8);
-			jp2_read_boxhdr(cinfo, cio, &box);
-		}
-	} while(JP2_COLR != box.type);
+	cinfo = jp2->cinfo;
 
 	jp2->meth = cio_read(cio, 1);		/* METH */
 	jp2->precedence = cio_read(cio, 1);	/* PRECEDENCE */
 	jp2->approx = cio_read(cio, 1);		/* APPROX */
 
-	if (jp2->meth == 1) {
-		jp2->enumcs = cio_read(cio, 4);	/* EnumCS */
-	} else {
-		/* skip PROFILE */
-		skip_len = box.init_pos + box.length - cio_tell(cio);
-		if (skip_len < 0) {
-			opj_event_msg(cinfo, EVT_ERROR, "Error with JP2H box size\n");
-			return false;
-		}
-		cio_skip(cio, box.init_pos + box.length - cio_tell(cio));
-	}
+	if (jp2->meth == 1) 
+   {
+	jp2->enumcs = cio_read(cio, 4);	/* EnumCS */
+   } 
+	else
+	if (jp2->meth == 2) 
+   {
+/* skip PROFILE */
+	skip_len = box->init_pos + box->length - cio_tell(cio);
+	if (skip_len < 0) 
+  {
+	opj_event_msg(cinfo, EVT_ERROR, "Error with COLR box size\n");
+	return OPJ_FALSE;
+  }
+	if(skip_len > 0)
+  {
+	unsigned char *start;
+
+	start = cio_getbp(cio);
+	color->icc_profile_buf = (unsigned char*)opj_malloc(skip_len);
+	color->icc_profile_len = skip_len;
+
+	cio_skip(cio, box->init_pos + box->length - cio_tell(cio));
+
+	memcpy(color->icc_profile_buf, start, skip_len);
+  }
+   }
+
+	if (cio_tell(cio) - box->init_pos != box->length) 
+   {
+	opj_event_msg(cinfo, EVT_ERROR, "Error with COLR Box\n");
+	return OPJ_FALSE;
+   }
+	color->jp2_has_colr = 1;
+
+	return OPJ_TRUE;
+}/* jp2_read_colr() */
+
+opj_bool jp2_read_jp2h(opj_jp2_t *jp2, opj_cio_t *cio, opj_jp2_color_t *color) 
+{
+	opj_jp2_box_t box;
+	int jp2h_end;
+
+	opj_common_ptr cinfo = jp2->cinfo;
+
+	jp2_read_boxhdr(cinfo, cio, &box);
+	do 
+   {
+	if (JP2_JP2H != box.type) 
+  {
+	if (box.type == JP2_JP2C) 
+ {
+	opj_event_msg(cinfo, EVT_ERROR, "Expected JP2H Marker\n");
+	return OPJ_FALSE;
+ }
+	cio_skip(cio, box.length - 8);
+
+	if(cio->bp >= cio->end) return OPJ_FALSE;
+
+	jp2_read_boxhdr(cinfo, cio, &box);
+  }
+   } while(JP2_JP2H != box.type);
+
+	if (!jp2_read_ihdr(jp2, cio))
+		return OPJ_FALSE;
+	jp2h_end = box.init_pos + box.length;
+
+	if (jp2->bpc == 255) 
+   {
+	if (!jp2_read_bpcc(jp2, cio))
+		return OPJ_FALSE;
+   }
+	jp2_read_boxhdr(cinfo, cio, &box);
+
+	while(cio_tell(cio) < jp2h_end)
+   {
+	if(box.type == JP2_COLR)
+  {
+	if( !jp2_read_colr(jp2, cio, &box, color))
+ {
+    cio_seek(cio, box.init_pos + 8);
+    cio_skip(cio, box.length - 8);
+ }
+    jp2_read_boxhdr(cinfo, cio, &box);
+    continue;
+  }
+    if(box.type == JP2_CDEF)
+  {
+    if( !jp2_read_cdef(jp2, cio, &box, color))
+ {
+    cio_seek(cio, box.init_pos + 8);
+    cio_skip(cio, box.length - 8);
+ }
+    jp2_read_boxhdr(cinfo, cio, &box);
+    continue;
+  }
+    if(box.type == JP2_PCLR)
+  {
+    if( !jp2_read_pclr(jp2, cio, &box, color))
+ {
+    cio_seek(cio, box.init_pos + 8);
+    cio_skip(cio, box.length - 8);
+ }
+    jp2_read_boxhdr(cinfo, cio, &box);
+    continue;
+  }
+    if(box.type == JP2_CMAP)
+  {
+    if( !jp2_read_cmap(jp2, cio, &box, color))
+ {
+    cio_seek(cio, box.init_pos + 8);
+    cio_skip(cio, box.length - 8);
+ }
+    jp2_read_boxhdr(cinfo, cio, &box);
+    continue;
+  }
+	cio_seek(cio, box.init_pos + 8);
+	cio_skip(cio, box.length - 8);
+	jp2_read_boxhdr(cinfo, cio, &box);
+
+   }/* while(cio_tell(cio) < box_end) */
+
+	cio_seek(cio, jp2h_end);
+
+/* Part 1, I.5.3.3 : 'must contain at least one' */
+	return (color->jp2_has_colr == 1);
+
+}/* jp2_read_jp2h() */
+
+opj_image_t* opj_jp2_decode(opj_jp2_t *jp2, opj_cio_t *cio, 
+	opj_codestream_info_t *cstr_info) 
+{
+	opj_common_ptr cinfo;
+	opj_image_t *image = NULL;
+	opj_jp2_color_t color;
+
+	if(!jp2 || !cio) 
+   {
+	return NULL;
+   }
+	memset(&color, 0, sizeof(opj_jp2_color_t));
+	cinfo = jp2->cinfo;
+
+/* JP2 decoding */
+	if(!jp2_read_struct(jp2, cio, &color)) 
+   {
+	free_color_data(&color);
+	opj_event_msg(cinfo, EVT_ERROR, "Failed to decode jp2 structure\n");
+	return NULL;
+   }
+
+/* J2K decoding */
+	image = j2k_decode(jp2->j2k, cio, cstr_info);
+
+	if(!image) 
+   {
+	free_color_data(&color);
+	opj_event_msg(cinfo, EVT_ERROR, "Failed to decode J2K image\n");
+	return NULL;
+   }
+
+/* Set Image Color Space */
+	if (jp2->enumcs == 16)
+		image->color_space = CLRSPC_SRGB;
+	else if (jp2->enumcs == 17)
+		image->color_space = CLRSPC_GRAY;
+	else if (jp2->enumcs == 18)
+		image->color_space = CLRSPC_SYCC;
+	else
+		image->color_space = CLRSPC_UNKNOWN;
+
+	if(color.jp2_cdef)
+   {
+	jp2_apply_cdef(image, &color);
+   }
+	if(color.jp2_pclr)
+   {
+/* Part 1, I.5.3.4: Either both or none : */
+	if( !color.jp2_pclr->cmap) 
+	 jp2_free_pclr(&color);
+	else
+	 jp2_apply_pclr(&color, image);
+   }
+	if(color.icc_profile_buf)
+   {
+	image->icc_profile_buf = color.icc_profile_buf;
+	color.icc_profile_buf = NULL;
+	image->icc_profile_len = color.icc_profile_len;
+   }
+	return image;
+
+}/* opj_jp2_decode() */
 
-	if (cio_tell(cio) - box.init_pos != box.length) {
-		opj_event_msg(cinfo, EVT_ERROR, "Error with BPCC Box\n");
-		return false;
-	}
-	return true;
-}
 
 void jp2_write_jp2h(opj_jp2_t *jp2, opj_cio_t *cio) {
 	opj_jp2_box_t box;
@@ -315,44 +766,6 @@ void jp2_write_jp2h(opj_jp2_t *jp2, opj_cio_t *cio) {
 	cio_seek(cio, box.init_pos + box.length);
 }
 
-bool jp2_read_jp2h(opj_jp2_t *jp2, opj_cio_t *cio) {
-	opj_jp2_box_t box;
-	int skip_len;
-
-	opj_common_ptr cinfo = jp2->cinfo;
-
-	jp2_read_boxhdr(cinfo, cio, &box);
-	do {
-		if (JP2_JP2H != box.type) {
-			if (box.type == JP2_JP2C) {
-				opj_event_msg(cinfo, EVT_ERROR, "Expected JP2H Marker\n");
-				return false;
-			}
-			cio_skip(cio, box.length - 8);
-			jp2_read_boxhdr(cinfo, cio, &box);
-		}
-	} while(JP2_JP2H != box.type);
-
-	if (!jp2_read_ihdr(jp2, cio))
-		return false;
-
-	if (jp2->bpc == 255) {
-		if (!jp2_read_bpcc(jp2, cio))
-			return false;
-	}
-	if (!jp2_read_colr(jp2, cio))
-		return false;
-
-	skip_len = box.init_pos + box.length - cio_tell(cio);
-	if (skip_len < 0) {
-		opj_event_msg(cinfo, EVT_ERROR, "Error with JP2H Box\n");
-		return false;
-	}
-	cio_skip(cio, box.init_pos + box.length - cio_tell(cio));
-
-	return true;
-}
-
 static void jp2_write_ftyp(opj_jp2_t *jp2, opj_cio_t *cio) {
 	unsigned int i;
 	opj_jp2_box_t box;
@@ -374,7 +787,7 @@ static void jp2_write_ftyp(opj_jp2_t *jp2, opj_cio_t *cio) {
 	cio_seek(cio, box.init_pos + box.length);
 }
 
-static bool jp2_read_ftyp(opj_jp2_t *jp2, opj_cio_t *cio) {
+static opj_bool jp2_read_ftyp(opj_jp2_t *jp2, opj_cio_t *cio) {
 	int i;
 	opj_jp2_box_t box;
 
@@ -384,7 +797,7 @@ static bool jp2_read_ftyp(opj_jp2_t *jp2, opj_cio_t *cio) {
 
 	if (JP2_FTYP != box.type) {
 		opj_event_msg(cinfo, EVT_ERROR, "Expected FTYP Marker\n");
-		return false;
+		return OPJ_FALSE;
 	}
 
 	jp2->brand = cio_read(cio, 4);		/* BR */
@@ -398,13 +811,13 @@ static bool jp2_read_ftyp(opj_jp2_t *jp2, opj_cio_t *cio) {
 
 	if (cio_tell(cio) - box.init_pos != box.length) {
 		opj_event_msg(cinfo, EVT_ERROR, "Error with FTYP Box\n");
-		return false;
+		return OPJ_FALSE;
 	}
 
-	return true;
+	return OPJ_TRUE;
 }
 
-static int jp2_write_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, opj_image_t *image, char *index) {
+static int jp2_write_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, opj_image_t *image, opj_codestream_info_t *cstr_info) {
 	unsigned int j2k_codestream_offset, j2k_codestream_length;
 	opj_jp2_box_t box;
 
@@ -416,7 +829,7 @@ static int jp2_write_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, opj_image_t *image, ch
 
 	/* J2K encoding */
 	j2k_codestream_offset = cio_tell(cio);
-	if(!j2k_encode(j2k, cio, image, index)) {
+	if(!j2k_encode(j2k, cio, image, cstr_info)) {
 		opj_event_msg(j2k->cinfo, EVT_ERROR, "Failed to encode image\n");
 		return 0;
 	}
@@ -433,7 +846,7 @@ static int jp2_write_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, opj_image_t *image, ch
 	return box.length;
 }
 
-static bool jp2_read_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, unsigned int *j2k_codestream_length, unsigned int *j2k_codestream_offset) {
+static opj_bool jp2_read_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, unsigned int *j2k_codestream_length, unsigned int *j2k_codestream_offset) {
 	opj_jp2_box_t box;
 
 	opj_common_ptr cinfo = jp2->cinfo;
@@ -449,7 +862,7 @@ static bool jp2_read_jp2c(opj_jp2_t *jp2, opj_cio_t *cio, unsigned int *j2k_code
 	*j2k_codestream_offset = cio_tell(cio);
 	*j2k_codestream_length = box.length - 8;
 
-	return true;
+	return OPJ_TRUE;
 }
 
 static void jp2_write_jp(opj_cio_t *cio) {
@@ -466,7 +879,7 @@ static void jp2_write_jp(opj_cio_t *cio) {
 	cio_seek(cio, box.init_pos + box.length);
 }
 
-static bool jp2_read_jp(opj_jp2_t *jp2, opj_cio_t *cio) {
+static opj_bool jp2_read_jp(opj_jp2_t *jp2, opj_cio_t *cio) {
 	opj_jp2_box_t box;
 
 	opj_common_ptr cinfo = jp2->cinfo;
@@ -474,32 +887,33 @@ static bool jp2_read_jp(opj_jp2_t *jp2, opj_cio_t *cio) {
 	jp2_read_boxhdr(cinfo, cio, &box);
 	if (JP2_JP != box.type) {
 		opj_event_msg(cinfo, EVT_ERROR, "Expected JP Marker\n");
-		return false;
+		return OPJ_FALSE;
 	}
 	if (0x0d0a870a != cio_read(cio, 4)) {
 		opj_event_msg(cinfo, EVT_ERROR, "Error with JP Marker\n");
-		return false;
+		return OPJ_FALSE;
 	}
 	if (cio_tell(cio) - box.init_pos != box.length) {
 		opj_event_msg(cinfo, EVT_ERROR, "Error with JP Box size\n");
-		return false;
+		return OPJ_FALSE;
 	}
 
-	return true;
+	return OPJ_TRUE;
 }
 
 
-static bool jp2_read_struct(opj_jp2_t *jp2, opj_cio_t *cio) {
+static opj_bool jp2_read_struct(opj_jp2_t *jp2, opj_cio_t *cio,
+	opj_jp2_color_t *color) {
 	if (!jp2_read_jp(jp2, cio))
-		return false;
+		return OPJ_FALSE;
 	if (!jp2_read_ftyp(jp2, cio))
-		return false;
-	if (!jp2_read_jp2h(jp2, cio))
-		return false;
+		return OPJ_FALSE;
+	if (!jp2_read_jp2h(jp2, cio, color))
+		return OPJ_FALSE;
 	if (!jp2_read_jp2c(jp2, cio, &jp2->j2k_codestream_length, &jp2->j2k_codestream_offset))
-		return false;
+		return OPJ_FALSE;
 	
-	return true;
+	return OPJ_TRUE;
 }
 
 /* ----------------------------------------------------------------------- */
@@ -507,7 +921,7 @@ static bool jp2_read_struct(opj_jp2_t *jp2, opj_cio_t *cio) {
 /* ----------------------------------------------------------------------- */
 
 opj_jp2_t* jp2_create_decompress(opj_common_ptr cinfo) {
-	opj_jp2_t *jp2 = (opj_jp2_t*)opj_malloc(sizeof(opj_jp2_t));
+	opj_jp2_t *jp2 = (opj_jp2_t*) opj_calloc(1, sizeof(opj_jp2_t));
 	if(jp2) {
 		jp2->cinfo = cinfo;
 		/* create the J2K codec */
@@ -541,41 +955,6 @@ void jp2_setup_decoder(opj_jp2_t *jp2, opj_dparameters_t *parameters) {
 	/* further JP2 initializations go here */
 }
 
-opj_image_t* jp2_decode(opj_jp2_t *jp2, opj_cio_t *cio) {
-	opj_common_ptr cinfo;
-	opj_image_t *image = NULL;
-
-	if(!jp2 || !cio) {
-		return NULL;
-	}
-
-	cinfo = jp2->cinfo;
-
-	/* JP2 decoding */
-	if(!jp2_read_struct(jp2, cio)) {
-		opj_event_msg(cinfo, EVT_ERROR, "Failed to decode jp2 structure\n");
-		return NULL;
-	}
-
-	/* J2K decoding */
-	image = j2k_decode(jp2->j2k, cio);
-	if(!image) {
-		opj_event_msg(cinfo, EVT_ERROR, "Failed to decode J2K image\n");
-	}
-
-	/* Set Image Color Space */
-	if (jp2->enumcs == 16)
-		image->color_space = CLRSPC_SRGB;
-	else if (jp2->enumcs == 17)
-		image->color_space = CLRSPC_GRAY;
-	else if (jp2->enumcs == 18)
-		image->color_space = CLRSPC_SYCC;
-	else
-		image->color_space = CLRSPC_UNKNOWN;
-
-	return image;
-}
-
 /* ----------------------------------------------------------------------- */
 /* JP2 encoder interface                                             */
 /* ----------------------------------------------------------------------- */
@@ -663,30 +1042,19 @@ void jp2_setup_encoder(opj_jp2_t *jp2, opj_cparameters_t *parameters, opj_image_
 	for (i = 0; i < image->numcomps; i++) {
 		jp2->comps[i].bpcc = image->comps[i].prec - 1 + (image->comps[i].sgnd << 7);
 	}
-
-	/* Colour Specification box */
-
-	if ((image->numcomps == 1 || image->numcomps == 3) && (jp2->bpc != 255)) {
-		jp2->meth = 1;	/* METH: Enumerated colourspace */
-	} else {
-		jp2->meth = 2;	/* METH: Restricted ICC profile */
-	}
-	if (jp2->meth == 1) {
-		if (image->color_space == 1)
-			jp2->enumcs = 16;	/* sRGB as defined by IEC 6196621 */
-		else if (image->color_space == 2)
-			jp2->enumcs = 17;	/* greyscale */
-		else if (image->color_space == 3)
-			jp2->enumcs = 18;	/* YUV */
-	} else {
-		jp2->enumcs = 0;		/* PROFILE (??) */
-	}
+	jp2->meth = 1;
+	if (image->color_space == 1)
+		jp2->enumcs = 16;	/* sRGB as defined by IEC 61966-2.1 */
+	else if (image->color_space == 2)
+		jp2->enumcs = 17;	/* greyscale */
+	else if (image->color_space == 3)
+		jp2->enumcs = 18;	/* YUV */
 	jp2->precedence = 0;	/* PRECEDENCE */
 	jp2->approx = 0;		/* APPROX */
 
 }
 
-bool jp2_encode(opj_jp2_t *jp2, opj_cio_t *cio, opj_image_t *image, char *index) {
+opj_bool opj_jp2_encode(opj_jp2_t *jp2, opj_cio_t *cio, opj_image_t *image, opj_codestream_info_t *cstr_info) {
 
 	/* JP2 encoding */
 
@@ -699,12 +1067,12 @@ bool jp2_encode(opj_jp2_t *jp2, opj_cio_t *cio, opj_image_t *image, char *index)
 
 	/* J2K encoding */
 
-	if(!jp2_write_jp2c(jp2, cio, image, index)) {
+	if(!jp2_write_jp2c(jp2, cio, image, cstr_info)) {
 		opj_event_msg(jp2->cinfo, EVT_ERROR, "Failed to encode image\n");
-		return false;
+		return OPJ_FALSE;
 	}
 
-	return true;
+	return OPJ_TRUE;
 }
 
 
