@@ -39,9 +39,9 @@
 /** @name Local static functions */
 /*@{*/
 
-static char t1_getctxno_zc(int f, int orient);
+static INLINE char t1_getctxno_zc(int f, int orient);
 static char t1_getctxno_sc(int f);
-static char t1_getctxno_mag(int f);
+static INLINE int t1_getctxno_mag(int f);
 static char t1_getspb(int f);
 static short t1_getnmsedec_sig(int x, int bitpos);
 static short t1_getnmsedec_ref(int x, int bitpos);
@@ -62,13 +62,25 @@ static void t1_enc_sigpass_step(
 /**
 Decode significant pass
 */
-static void t1_dec_sigpass_step(
+static INLINE void t1_dec_sigpass_step_raw(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int orient,
+		int oneplushalf,
+		int vsc);
+static INLINE void t1_dec_sigpass_step_mqc(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int orient,
+		int oneplushalf);
+static INLINE void t1_dec_sigpass_step_mqc_vsc(
 		opj_t1_t *t1,
 		flag_t *flagsp,
 		int *datap,
 		int orient,
 		int oneplushalf,
-		char type,
 		int vsc);
 /**
 Encode significant pass
@@ -83,12 +95,19 @@ static void t1_enc_sigpass(
 /**
 Decode significant pass
 */
-static void t1_dec_sigpass(
+static void t1_dec_sigpass_raw(
 		opj_t1_t *t1,
 		int bpno,
 		int orient,
-		char type,
 		int cblksty);
+static void t1_dec_sigpass_mqc(
+		opj_t1_t *t1,
+		int bpno,
+		int orient);
+static void t1_dec_sigpass_mqc_vsc(
+		opj_t1_t *t1,
+		int bpno,
+		int orient);
 /**
 Encode refinement pass
 */
@@ -104,14 +123,27 @@ static void t1_enc_refpass_step(
 /**
 Decode refinement pass
 */
-static void t1_dec_refpass_step(
+static void INLINE t1_dec_refpass_step_raw(
 		opj_t1_t *t1,
 		flag_t *flagsp,
 		int *datap,
 		int poshalf,
 		int neghalf,
-		char type,
 		int vsc);
+static void INLINE t1_dec_refpass_step_mqc(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int poshalf,
+		int neghalf);
+static void INLINE t1_dec_refpass_step_mqc_vsc(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int poshalf,
+		int neghalf,
+		int vsc);
+
 /**
 Encode refinement pass
 */
@@ -124,11 +156,16 @@ static void t1_enc_refpass(
 /**
 Decode refinement pass
 */
-static void t1_dec_refpass(
+static void t1_dec_refpass_raw(
 		opj_t1_t *t1,
 		int bpno,
-		char type,
 		int cblksty);
+static void t1_dec_refpass_mqc(
+		opj_t1_t *t1,
+		int bpno);
+static void t1_dec_refpass_mqc_vsc(
+		opj_t1_t *t1,
+		int bpno);
 /**
 Encode clean-up pass
 */
@@ -145,7 +182,19 @@ static void t1_enc_clnpass_step(
 /**
 Decode clean-up pass
 */
+static void t1_dec_clnpass_step_partial(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int orient,
+		int oneplushalf);
 static void t1_dec_clnpass_step(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int orient,
+		int oneplushalf);
+static void t1_dec_clnpass_step_vsc(
 		opj_t1_t *t1,
 		flag_t *flagsp,
 		int *datap,
@@ -178,7 +227,8 @@ static double t1_getwmsedec(
 		int bpno,
 		int qmfbid,
 		double stepsize,
-		int numcomps);
+		int numcomps,
+		int mct);
 /**
 Encode 1 code-block
 @param t1 T1 handle
@@ -190,11 +240,12 @@ Encode 1 code-block
 @param stepsize
 @param cblksty Code-block style
 @param numcomps
+@param mct
 @param tile
 */
 static void t1_encode_cblk(
 		opj_t1_t *t1,
-		opj_tcd_cblk_t * cblk,
+		opj_tcd_cblk_enc_t* cblk,
 		int orient,
 		int compno,
 		int level,
@@ -202,6 +253,7 @@ static void t1_encode_cblk(
 		double stepsize,
 		int cblksty,
 		int numcomps,
+		int mct,
 		opj_tcd_tile_t * tile);
 /**
 Decode 1 code-block
@@ -213,7 +265,7 @@ Decode 1 code-block
 */
 static void t1_decode_cblk(
 		opj_t1_t *t1,
-		opj_tcd_cblk_t * cblk,
+		opj_tcd_cblk_dec_t* cblk,
 		int orient,
 		int roishift,
 		int cblksty);
@@ -232,8 +284,10 @@ static char t1_getctxno_sc(int f) {
 	return lut_ctxno_sc[(f & (T1_SIG_PRIM | T1_SGN)) >> 4];
 }
 
-static char t1_getctxno_mag(int f) {
-	return lut_ctxno_mag[(f & T1_SIG_OTH) | (((f & T1_REFINE) != 0) << 11)];
+static int t1_getctxno_mag(int f) {
+	int tmp1 = (f & T1_SIG_OTH) ? T1_CTXNO_MAG + 1 : T1_CTXNO_MAG;
+	int tmp2 = (f & T1_REFINE) ? T1_CTXNO_MAG + 2 : tmp1;
+	return (tmp2);
 }
 
 static char t1_getspb(int f) {
@@ -261,21 +315,22 @@ static void t1_updateflags(flag_t *flagsp, int s, int stride) {
 	flag_t *sp = flagsp + stride;
 
 	static const flag_t mod[] = {
-		T1_SIG_S,            T1_SIG_N,            T1_SIG_E,            T1_SIG_W,
-		T1_SIG_S | T1_SGN_S, T1_SIG_N | T1_SGN_N, T1_SIG_E | T1_SGN_E, T1_SIG_W | T1_SGN_W
+		T1_SIG_S, T1_SIG_S|T1_SGN_S,
+		T1_SIG_E, T1_SIG_E|T1_SGN_E,
+		T1_SIG_W, T1_SIG_W|T1_SGN_W,
+		T1_SIG_N, T1_SIG_N|T1_SGN_N
 	};
 
-	s <<= 2;
-
 	np[-1] |= T1_SIG_SE;
 	np[0]  |= mod[s];
 	np[1]  |= T1_SIG_SW;
 
 	flagsp[-1] |= mod[s+2];
-	flagsp[1]  |= mod[s+3];
+	flagsp[0]  |= T1_SIG;
+	flagsp[1]  |= mod[s+4];
 
 	sp[-1] |= T1_SIG_NE;
-	sp[0]  |= mod[s+1];
+	sp[0]  |= mod[s+6];
 	sp[1]  |= T1_SIG_NW;
 }
 
@@ -297,62 +352,94 @@ static void t1_enc_sigpass_step(
 	flag = vsc ? ((*flagsp) & (~(T1_SIG_S | T1_SIG_SE | T1_SIG_SW | T1_SGN_S))) : (*flagsp);
 	if ((flag & T1_SIG_OTH) && !(flag & (T1_SIG | T1_VISIT))) {
 		v = int_abs(*datap) & one ? 1 : 0;
+		mqc_setcurctx(mqc, t1_getctxno_zc(flag, orient));	/* ESSAI */
 		if (type == T1_TYPE_RAW) {	/* BYPASS/LAZY MODE */
-			mqc_setcurctx(mqc, t1_getctxno_zc(flag, orient));	/* ESSAI */
 			mqc_bypass_enc(mqc, v);
 		} else {
-			mqc_setcurctx(mqc, t1_getctxno_zc(flag, orient));
 			mqc_encode(mqc, v);
 		}
 		if (v) {
 			v = *datap < 0 ? 1 : 0;
 			*nmsedec +=	t1_getnmsedec_sig(int_abs(*datap), bpno + T1_NMSEDEC_FRACBITS);
+			mqc_setcurctx(mqc, t1_getctxno_sc(flag));	/* ESSAI */
 			if (type == T1_TYPE_RAW) {	/* BYPASS/LAZY MODE */
-				mqc_setcurctx(mqc, t1_getctxno_sc(flag));	/* ESSAI */
 				mqc_bypass_enc(mqc, v);
 			} else {
-				mqc_setcurctx(mqc, t1_getctxno_sc(flag));
 				mqc_encode(mqc, v ^ t1_getspb(flag));
 			}
 			t1_updateflags(flagsp, v, t1->flags_stride);
-			*flagsp |= T1_SIG;
 		}
 		*flagsp |= T1_VISIT;
 	}
 }
 
-static void t1_dec_sigpass_step(
+static INLINE void t1_dec_sigpass_step_raw(
 		opj_t1_t *t1,
 		flag_t *flagsp,
 		int *datap,
 		int orient,
 		int oneplushalf,
-		char type,
 		int vsc)
 {
 	int v, flag;
 	
 	opj_raw_t *raw = t1->raw;	/* RAW component */
-	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
 	
 	flag = vsc ? ((*flagsp) & (~(T1_SIG_S | T1_SIG_SE | T1_SIG_SW | T1_SGN_S))) : (*flagsp);
 	if ((flag & T1_SIG_OTH) && !(flag & (T1_SIG | T1_VISIT))) {
-		if (type == T1_TYPE_RAW) {
 			if (raw_decode(raw)) {
 				v = raw_decode(raw);	/* ESSAI */
 				*datap = v ? -oneplushalf : oneplushalf;
 				t1_updateflags(flagsp, v, t1->flags_stride);
-				*flagsp |= T1_SIG;
 			}
-		} else {
+		*flagsp |= T1_VISIT;
+	}
+}				/* VSC and  BYPASS by Antonin */
+
+static INLINE void t1_dec_sigpass_step_mqc(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int orient,
+		int oneplushalf)
+{
+	int v, flag;
+	
+	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
+	
+	flag = *flagsp;
+	if ((flag & T1_SIG_OTH) && !(flag & (T1_SIG | T1_VISIT))) {
 			mqc_setcurctx(mqc, t1_getctxno_zc(flag, orient));
 			if (mqc_decode(mqc)) {
 				mqc_setcurctx(mqc, t1_getctxno_sc(flag));
 				v = mqc_decode(mqc) ^ t1_getspb(flag);
 				*datap = v ? -oneplushalf : oneplushalf;
 				t1_updateflags(flagsp, v, t1->flags_stride);
-				*flagsp |= T1_SIG;
 			}
+		*flagsp |= T1_VISIT;
+	}
+}				/* VSC and  BYPASS by Antonin */
+
+static INLINE void t1_dec_sigpass_step_mqc_vsc(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int orient,
+		int oneplushalf,
+		int vsc)
+{
+	int v, flag;
+	
+	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
+	
+	flag = vsc ? ((*flagsp) & (~(T1_SIG_S | T1_SIG_SE | T1_SIG_SW | T1_SGN_S))) : (*flagsp);
+	if ((flag & T1_SIG_OTH) && !(flag & (T1_SIG | T1_VISIT))) {
+		mqc_setcurctx(mqc, t1_getctxno_zc(flag, orient));
+		if (mqc_decode(mqc)) {
+			mqc_setcurctx(mqc, t1_getctxno_sc(flag));
+			v = mqc_decode(mqc) ^ t1_getspb(flag);
+			*datap = v ? -oneplushalf : oneplushalf;
+			t1_updateflags(flagsp, v, t1->flags_stride);
 		}
 		*flagsp |= T1_VISIT;
 	}
@@ -388,11 +475,10 @@ static void t1_enc_sigpass(
 	}
 }
 
-static void t1_dec_sigpass(
+static void t1_dec_sigpass_raw(
 		opj_t1_t *t1,
 		int bpno,
 		int orient,
-		char type,
 		int cblksty)
 {
 	int i, j, k, one, half, oneplushalf, vsc;
@@ -403,13 +489,79 @@ static void t1_dec_sigpass(
 		for (i = 0; i < t1->w; ++i) {
 			for (j = k; j < k + 4 && j < t1->h; ++j) {
 				vsc = ((cblksty & J2K_CCP_CBLKSTY_VSC) && (j == k + 3 || j == t1->h - 1)) ? 1 : 0;
-				t1_dec_sigpass_step(
+				t1_dec_sigpass_step_raw(
+						t1,
+						&t1->flags[((j+1) * t1->flags_stride) + i + 1],
+						&t1->data[(j * t1->w) + i],
+						orient,
+						oneplushalf,
+						vsc);
+			}
+		}
+	}
+}				/* VSC and  BYPASS by Antonin */
+
+static void t1_dec_sigpass_mqc(
+		opj_t1_t *t1,
+		int bpno,
+		int orient)
+{
+	int i, j, k, one, half, oneplushalf;
+	int *data1 = t1->data;
+	flag_t *flags1 = &t1->flags[1];
+	one = 1 << bpno;
+	half = one >> 1;
+	oneplushalf = one | half;
+	for (k = 0; k < (t1->h & ~3); k += 4) {
+		for (i = 0; i < t1->w; ++i) {
+			int *data2 = data1 + i;
+			flag_t *flags2 = flags1 + i;
+			flags2 += t1->flags_stride;
+			t1_dec_sigpass_step_mqc(t1, flags2, data2, orient, oneplushalf);
+			data2 += t1->w;
+			flags2 += t1->flags_stride;
+			t1_dec_sigpass_step_mqc(t1, flags2, data2, orient, oneplushalf);
+			data2 += t1->w;
+			flags2 += t1->flags_stride;
+			t1_dec_sigpass_step_mqc(t1, flags2, data2, orient, oneplushalf);
+			data2 += t1->w;
+			flags2 += t1->flags_stride;
+			t1_dec_sigpass_step_mqc(t1, flags2, data2, orient, oneplushalf);
+			data2 += t1->w;
+		}
+		data1 += t1->w << 2;
+		flags1 += t1->flags_stride << 2;
+	}
+	for (i = 0; i < t1->w; ++i) {
+		int *data2 = data1 + i;
+		flag_t *flags2 = flags1 + i;
+		for (j = k; j < t1->h; ++j) {
+			flags2 += t1->flags_stride;
+			t1_dec_sigpass_step_mqc(t1, flags2, data2, orient, oneplushalf);
+			data2 += t1->w;
+		}
+	}
+}				/* VSC and  BYPASS by Antonin */
+
+static void t1_dec_sigpass_mqc_vsc(
+		opj_t1_t *t1,
+		int bpno,
+		int orient)
+{
+	int i, j, k, one, half, oneplushalf, vsc;
+	one = 1 << bpno;
+	half = one >> 1;
+	oneplushalf = one | half;
+	for (k = 0; k < t1->h; k += 4) {
+		for (i = 0; i < t1->w; ++i) {
+			for (j = k; j < k + 4 && j < t1->h; ++j) {
+				vsc = (j == k + 3 || j == t1->h - 1) ? 1 : 0;
+				t1_dec_sigpass_step_mqc_vsc(
 						t1,
 						&t1->flags[((j+1) * t1->flags_stride) + i + 1],
 						&t1->data[(j * t1->w) + i],
 						orient,
 						oneplushalf,
-						type,
 						vsc);
 			}
 		}
@@ -434,40 +586,74 @@ static void t1_enc_refpass_step(
 	if ((flag & (T1_SIG | T1_VISIT)) == T1_SIG) {
 		*nmsedec += t1_getnmsedec_ref(int_abs(*datap), bpno + T1_NMSEDEC_FRACBITS);
 		v = int_abs(*datap) & one ? 1 : 0;
+		mqc_setcurctx(mqc, t1_getctxno_mag(flag));	/* ESSAI */
 		if (type == T1_TYPE_RAW) {	/* BYPASS/LAZY MODE */
-			mqc_setcurctx(mqc, t1_getctxno_mag(flag));	/* ESSAI */
 			mqc_bypass_enc(mqc, v);
 		} else {
-			mqc_setcurctx(mqc, t1_getctxno_mag(flag));
 			mqc_encode(mqc, v);
 		}
 		*flagsp |= T1_REFINE;
 	}
 }
 
-static void t1_dec_refpass_step(
+static INLINE void t1_dec_refpass_step_raw(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int poshalf,
+		int neghalf,
+		int vsc)
+{
+	int v, t, flag;
+	
+	opj_raw_t *raw = t1->raw;	/* RAW component */
+	
+	flag = vsc ? ((*flagsp) & (~(T1_SIG_S | T1_SIG_SE | T1_SIG_SW | T1_SGN_S))) : (*flagsp);
+	if ((flag & (T1_SIG | T1_VISIT)) == T1_SIG) {
+			v = raw_decode(raw);
+		t = v ? poshalf : neghalf;
+		*datap += *datap < 0 ? -t : t;
+		*flagsp |= T1_REFINE;
+	}
+}				/* VSC and  BYPASS by Antonin  */
+
+static INLINE void t1_dec_refpass_step_mqc(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int poshalf,
+		int neghalf)
+{
+	int v, t, flag;
+	
+	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
+	
+	flag = *flagsp;
+	if ((flag & (T1_SIG | T1_VISIT)) == T1_SIG) {
+		mqc_setcurctx(mqc, t1_getctxno_mag(flag));	/* ESSAI */
+			v = mqc_decode(mqc);
+		t = v ? poshalf : neghalf;
+		*datap += *datap < 0 ? -t : t;
+		*flagsp |= T1_REFINE;
+		}
+}				/* VSC and  BYPASS by Antonin  */
+
+static INLINE void t1_dec_refpass_step_mqc_vsc(
 		opj_t1_t *t1,
 		flag_t *flagsp,
 		int *datap,
 		int poshalf,
 		int neghalf,
-		char type,
 		int vsc)
 {
 	int v, t, flag;
 	
 	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
-	opj_raw_t *raw = t1->raw;	/* RAW component */
 	
 	flag = vsc ? ((*flagsp) & (~(T1_SIG_S | T1_SIG_SE | T1_SIG_SW | T1_SGN_S))) : (*flagsp);
 	if ((flag & (T1_SIG | T1_VISIT)) == T1_SIG) {
-		if (type == T1_TYPE_RAW) {
-			mqc_setcurctx(mqc, t1_getctxno_mag(flag));	/* ESSAI */
-			v = raw_decode(raw);
-		} else {
-			mqc_setcurctx(mqc, t1_getctxno_mag(flag));
-			v = mqc_decode(mqc);
-		}
+		mqc_setcurctx(mqc, t1_getctxno_mag(flag));	/* ESSAI */
+		v = mqc_decode(mqc);
 		t = v ? poshalf : neghalf;
 		*datap += *datap < 0 ? -t : t;
 		*flagsp |= T1_REFINE;
@@ -502,10 +688,9 @@ static void t1_enc_refpass(
 	}
 }
 
-static void t1_dec_refpass(
+static void t1_dec_refpass_raw(
 		opj_t1_t *t1,
 		int bpno,
-		char type,
 		int cblksty)
 {
 	int i, j, k, one, poshalf, neghalf;
@@ -517,13 +702,78 @@ static void t1_dec_refpass(
 		for (i = 0; i < t1->w; ++i) {
 			for (j = k; j < k + 4 && j < t1->h; ++j) {
 				vsc = ((cblksty & J2K_CCP_CBLKSTY_VSC) && (j == k + 3 || j == t1->h - 1)) ? 1 : 0;
-				t1_dec_refpass_step(
+				t1_dec_refpass_step_raw(
+						t1,
+						&t1->flags[((j+1) * t1->flags_stride) + i + 1],
+						&t1->data[(j * t1->w) + i],
+						poshalf,
+						neghalf,
+						vsc);
+			}
+		}
+	}
+}				/* VSC and  BYPASS by Antonin */
+
+static void t1_dec_refpass_mqc(
+		opj_t1_t *t1,
+		int bpno)
+{
+	int i, j, k, one, poshalf, neghalf;
+	int *data1 = t1->data;
+	flag_t *flags1 = &t1->flags[1];
+	one = 1 << bpno;
+	poshalf = one >> 1;
+	neghalf = bpno > 0 ? -poshalf : -1;
+	for (k = 0; k < (t1->h & ~3); k += 4) {
+		for (i = 0; i < t1->w; ++i) {
+			int *data2 = data1 + i;
+			flag_t *flags2 = flags1 + i;
+			flags2 += t1->flags_stride;
+			t1_dec_refpass_step_mqc(t1, flags2, data2, poshalf, neghalf);
+			data2 += t1->w;
+			flags2 += t1->flags_stride;
+			t1_dec_refpass_step_mqc(t1, flags2, data2, poshalf, neghalf);
+			data2 += t1->w;
+			flags2 += t1->flags_stride;
+			t1_dec_refpass_step_mqc(t1, flags2, data2, poshalf, neghalf);
+			data2 += t1->w;
+			flags2 += t1->flags_stride;
+			t1_dec_refpass_step_mqc(t1, flags2, data2, poshalf, neghalf);
+			data2 += t1->w;
+		}
+		data1 += t1->w << 2;
+		flags1 += t1->flags_stride << 2;
+	}
+	for (i = 0; i < t1->w; ++i) {
+		int *data2 = data1 + i;
+		flag_t *flags2 = flags1 + i;
+		for (j = k; j < t1->h; ++j) {
+			flags2 += t1->flags_stride;
+			t1_dec_refpass_step_mqc(t1, flags2, data2, poshalf, neghalf);
+			data2 += t1->w;
+		}
+	}
+}				/* VSC and  BYPASS by Antonin */
+
+static void t1_dec_refpass_mqc_vsc(
+		opj_t1_t *t1,
+		int bpno)
+{
+	int i, j, k, one, poshalf, neghalf;
+	int vsc;
+	one = 1 << bpno;
+	poshalf = one >> 1;
+	neghalf = bpno > 0 ? -poshalf : -1;
+	for (k = 0; k < t1->h; k += 4) {
+		for (i = 0; i < t1->w; ++i) {
+			for (j = k; j < k + 4 && j < t1->h; ++j) {
+				vsc = ((j == k + 3 || j == t1->h - 1)) ? 1 : 0;
+				t1_dec_refpass_step_mqc_vsc(
 						t1,
 						&t1->flags[((j+1) * t1->flags_stride) + i + 1],
 						&t1->data[(j * t1->w) + i],
 						poshalf,
 						neghalf,
-						type,
 						vsc);
 			}
 		}
@@ -560,13 +810,55 @@ LABEL_PARTIAL:
 			v = *datap < 0 ? 1 : 0;
 			mqc_encode(mqc, v ^ t1_getspb(flag));
 			t1_updateflags(flagsp, v, t1->flags_stride);
-			*flagsp |= T1_SIG;
 		}
 	}
 	*flagsp &= ~T1_VISIT;
 }
 
+static void t1_dec_clnpass_step_partial(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int orient,
+		int oneplushalf)
+{
+	int v, flag;
+	
+	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
+	
+	flag = *flagsp;
+	mqc_setcurctx(mqc, t1_getctxno_sc(flag));
+	v = mqc_decode(mqc) ^ t1_getspb(flag);
+	*datap = v ? -oneplushalf : oneplushalf;
+	t1_updateflags(flagsp, v, t1->flags_stride);
+	*flagsp &= ~T1_VISIT;
+}				/* VSC and  BYPASS by Antonin */
+
 static void t1_dec_clnpass_step(
+		opj_t1_t *t1,
+		flag_t *flagsp,
+		int *datap,
+		int orient,
+		int oneplushalf)
+{
+	int v, flag;
+	
+	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
+	
+	flag = *flagsp;
+	if (!(flag & (T1_SIG | T1_VISIT))) {
+		mqc_setcurctx(mqc, t1_getctxno_zc(flag, orient));
+		if (mqc_decode(mqc)) {
+			mqc_setcurctx(mqc, t1_getctxno_sc(flag));
+			v = mqc_decode(mqc) ^ t1_getspb(flag);
+			*datap = v ? -oneplushalf : oneplushalf;
+			t1_updateflags(flagsp, v, t1->flags_stride);
+		}
+	}
+	*flagsp &= ~T1_VISIT;
+}				/* VSC and  BYPASS by Antonin */
+
+static void t1_dec_clnpass_step_vsc(
 		opj_t1_t *t1,
 		flag_t *flagsp,
 		int *datap,
@@ -591,11 +883,10 @@ LABEL_PARTIAL:
 			v = mqc_decode(mqc) ^ t1_getspb(flag);
 			*datap = v ? -oneplushalf : oneplushalf;
 			t1_updateflags(flagsp, v, t1->flags_stride);
-			*flagsp |= T1_SIG;
 		}
 	}
 	*flagsp &= ~T1_VISIT;
-}				/* VSC and  BYPASS by Antonin */
+}
 
 static void t1_enc_clnpass(
 		opj_t1_t *t1,
@@ -675,22 +966,16 @@ static void t1_dec_clnpass(
 	one = 1 << bpno;
 	half = one >> 1;
 	oneplushalf = one | half;
+	if (cblksty & J2K_CCP_CBLKSTY_VSC) {
 	for (k = 0; k < t1->h; k += 4) {
 		for (i = 0; i < t1->w; ++i) {
 			if (k + 3 < t1->h) {
-				if (cblksty & J2K_CCP_CBLKSTY_VSC) {
 					agg = !(MACRO_t1_flags(1 + k,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH)
 						|| MACRO_t1_flags(1 + k + 1,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH)
 						|| MACRO_t1_flags(1 + k + 2,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH)
 						|| (MACRO_t1_flags(1 + k + 3,1 + i) 
 						& (~(T1_SIG_S | T1_SIG_SE | T1_SIG_SW |	T1_SGN_S))) & (T1_SIG | T1_VISIT | T1_SIG_OTH));
 				} else {
-					agg = !(MACRO_t1_flags(1 + k,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH)
-						|| MACRO_t1_flags(1 + k + 1,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH)
-						|| MACRO_t1_flags(1 + k + 2,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH)
-						|| MACRO_t1_flags(1 + k + 3,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH));
-				}
-			} else {
 				agg = 0;
 			}
 			if (agg) {
@@ -705,8 +990,8 @@ static void t1_dec_clnpass(
 				runlen = 0;
 			}
 			for (j = k + runlen; j < k + 4 && j < t1->h; ++j) {
-				vsc = ((cblksty & J2K_CCP_CBLKSTY_VSC) && (j == k + 3 || j == t1->h - 1)) ? 1 : 0;
-				t1_dec_clnpass_step(
+					vsc = (j == k + 3 || j == t1->h - 1) ? 1 : 0;
+					t1_dec_clnpass_step_vsc(
 						t1,
 						&t1->flags[((j+1) * t1->flags_stride) + i + 1],
 						&t1->data[(j * t1->w) + i],
@@ -717,6 +1002,65 @@ static void t1_dec_clnpass(
 			}
 		}
 	}
+	} else {
+		int *data1 = t1->data;
+		flag_t *flags1 = &t1->flags[1];
+		for (k = 0; k < (t1->h & ~3); k += 4) {
+			for (i = 0; i < t1->w; ++i) {
+				int *data2 = data1 + i;
+				flag_t *flags2 = flags1 + i;
+				agg = !(MACRO_t1_flags(1 + k,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH)
+					|| MACRO_t1_flags(1 + k + 1,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH)
+					|| MACRO_t1_flags(1 + k + 2,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH)
+					|| MACRO_t1_flags(1 + k + 3,1 + i) & (T1_SIG | T1_VISIT | T1_SIG_OTH));
+				if (agg) {
+					mqc_setcurctx(mqc, T1_CTXNO_AGG);
+					if (!mqc_decode(mqc)) {
+						continue;
+					}
+					mqc_setcurctx(mqc, T1_CTXNO_UNI);
+					runlen = mqc_decode(mqc);
+					runlen = (runlen << 1) | mqc_decode(mqc);
+					flags2 += runlen * t1->flags_stride;
+					data2 += runlen * t1->w;
+					for (j = k + runlen; j < k + 4 && j < t1->h; ++j) {
+						flags2 += t1->flags_stride;
+						if (agg && (j == k + runlen)) {
+							t1_dec_clnpass_step_partial(t1, flags2, data2, orient, oneplushalf);
+						} else {
+							t1_dec_clnpass_step(t1, flags2, data2, orient, oneplushalf);
+						}
+						data2 += t1->w;
+					}
+				} else {
+					flags2 += t1->flags_stride;
+					t1_dec_clnpass_step(t1, flags2, data2, orient, oneplushalf);
+					data2 += t1->w;
+					flags2 += t1->flags_stride;
+					t1_dec_clnpass_step(t1, flags2, data2, orient, oneplushalf);
+					data2 += t1->w;
+					flags2 += t1->flags_stride;
+					t1_dec_clnpass_step(t1, flags2, data2, orient, oneplushalf);
+					data2 += t1->w;
+					flags2 += t1->flags_stride;
+					t1_dec_clnpass_step(t1, flags2, data2, orient, oneplushalf);
+					data2 += t1->w;
+				}
+			}
+			data1 += t1->w << 2;
+			flags1 += t1->flags_stride << 2;
+		}
+		for (i = 0; i < t1->w; ++i) {
+			int *data2 = data1 + i;
+			flag_t *flags2 = flags1 + i;
+			for (j = k; j < t1->h; ++j) {
+				flags2 += t1->flags_stride;
+				t1_dec_clnpass_step(t1, flags2, data2, orient, oneplushalf);
+				data2 += t1->w;
+			}
+		}
+	}
+
 	if (segsym) {
 		int v = 0;
 		mqc_setcurctx(mqc, T1_CTXNO_UNI);
@@ -742,14 +1086,15 @@ static double t1_getwmsedec(
 		int bpno,
 		int qmfbid,
 		double stepsize,
-		int numcomps)
+		int numcomps,
+		int mct)
 {
 	double w1, w2, wmsedec;
 	if (qmfbid == 1) {
-		w1 = (numcomps > 1) ? mct_getnorm(compno) : 1.0;
+		w1 = (mct && numcomps==3) ? mct_getnorm(compno) : 1.0;
 		w2 = dwt_getnorm(level, orient);
 	} else {			/* if (qmfbid == 0) */
-		w1 = (numcomps > 1) ? mct_getnorm_real(compno) : 1.0;
+		w1 = (mct && numcomps==3) ? mct_getnorm_real(compno) : 1.0;
 		w2 = dwt_getnorm_real(level, orient);
 	}
 	wmsedec = w1 * w2 * stepsize * (1 << bpno);
@@ -758,53 +1103,47 @@ static double t1_getwmsedec(
 	return wmsedec;
 }
 
-static void allocate_buffers(
+static opj_bool allocate_buffers(
 		opj_t1_t *t1,
 		int w,
 		int h)
 {
-	int datasize;
+	int datasize=w * h;
 	int flagssize;
 
-	datasize=w * h;
-	//fprintf(stderr,"w=%i h=%i datasize=%i flagssize=%i\n",w,h,datasize,flagssize);
-
 	if(datasize > t1->datasize){
-		//fprintf(stderr,"Allocating t1->data: datasize=%i\n",datasize);
-		free(t1->data);
-		t1->data=malloc(datasize * sizeof(int));
+		opj_aligned_free(t1->data);
+		t1->data = (int*) opj_aligned_malloc(datasize * sizeof(int));
 		if(!t1->data){
-			return;
+			return OPJ_FALSE;
 		}
 		t1->datasize=datasize;
 	}
-	//memset(t1->data,0xff,t1->datasize);
 	memset(t1->data,0,datasize * sizeof(int));
 
 	t1->flags_stride=w+2;
 	flagssize=t1->flags_stride * (h+2);
 
 	if(flagssize > t1->flagssize){
-		//fprintf(stderr,"Allocating t1->flags: flagssize=%i\n",flagssize);
-		free(t1->flags);
-		t1->flags=malloc(flagssize * sizeof(flag_t));
+		opj_aligned_free(t1->flags);
+		t1->flags = (flag_t*) opj_aligned_malloc(flagssize * sizeof(flag_t));
 		if(!t1->flags){
-			fprintf(stderr,"Allocating t1->flags FAILED!\n");
-			return;
+			return OPJ_FALSE;
 		}
 		t1->flagssize=flagssize;
 	}
-	//memset(t1->flags,0xff,t1->flagssize);
 	memset(t1->flags,0,flagssize * sizeof(flag_t));
 
 	t1->w=w;
 	t1->h=h;
+
+	return OPJ_TRUE;
 }
 
 /** mod fixed_quality */
 static void t1_encode_cblk(
 		opj_t1_t *t1,
-		opj_tcd_cblk_t * cblk,
+		opj_tcd_cblk_enc_t* cblk,
 		int orient,
 		int compno,
 		int level,
@@ -812,25 +1151,25 @@ static void t1_encode_cblk(
 		double stepsize,
 		int cblksty,
 		int numcomps,
+		int mct,
 		opj_tcd_tile_t * tile)
 {
-	int i, j;
-	int passno;
-	int bpno, passtype;
-	int max;
-	int nmsedec = 0;
 	double cumwmsedec = 0.0;
-	char type = T1_TYPE_MQ;
-	
+
 	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
-	
+
+	int passno, bpno, passtype;
+	int nmsedec = 0;
+	int i, max;
+	char type = T1_TYPE_MQ;
+	double tempwmsedec;
+
 	max = 0;
-	for (j = 0; j < t1->h; ++j) {
-		for (i = 0; i < t1->w; ++i) {
-			max = int_max(max, int_abs(t1->data[(j * t1->w) + i]));
-		}
+	for (i = 0; i < t1->w * t1->h; ++i) {
+		int tmp = abs(t1->data[i]);
+		max = int_max(max, tmp);
 	}
-	
+
 	cblk->numbps = max ? (int_floorlog2(max) + 1) - T1_NMSEDEC_FRACBITS : 0;
 	
 	bpno = cblk->numbps - 1;
@@ -863,8 +1202,9 @@ static void t1_encode_cblk(
 		}
 		
 		/* fixed_quality */
-		cumwmsedec += t1_getwmsedec(nmsedec, compno, level, orient, bpno, qmfbid, stepsize, numcomps);
-		tile->distotile += t1_getwmsedec(nmsedec, compno, level, orient, bpno, qmfbid, stepsize, numcomps);
+		tempwmsedec = t1_getwmsedec(nmsedec, compno, level, orient, bpno, qmfbid, stepsize, numcomps, mct);
+		cumwmsedec += tempwmsedec;
+		tile->distotile += tempwmsedec;
 		
 		/* Code switch "RESTART" (i.e. TERMALL) */
 		if ((cblksty & J2K_CCP_CBLKSTY_TERMALL)	&& !((passtype == 2) && (bpno - 1 < 0))) {
@@ -937,23 +1277,26 @@ static void t1_encode_cblk(
 
 static void t1_decode_cblk(
 		opj_t1_t *t1,
-		opj_tcd_cblk_t * cblk,
+		opj_tcd_cblk_dec_t* cblk,
 		int orient,
 		int roishift,
 		int cblksty)
 {
+	opj_raw_t *raw = t1->raw;	/* RAW component */
+	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
+
 	int bpno, passtype;
 	int segno, passno;
 	char type = T1_TYPE_MQ; /* BYPASS mode */
-	
-	opj_raw_t *raw = t1->raw;	/* RAW component */
-	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
-	
-	allocate_buffers(
-			t1,
-			cblk->x1 - cblk->x0,
-			cblk->y1 - cblk->y0);
-	
+
+	if(!allocate_buffers(
+				t1,
+				cblk->x1 - cblk->x0,
+				cblk->y1 - cblk->y0))
+	{
+		return;
+	}
+
 	bpno = roishift + cblk->numbps - 1;
 	passtype = 2;
 	
@@ -967,19 +1310,39 @@ static void t1_decode_cblk(
 		
 		/* BYPASS mode */
 		type = ((bpno <= (cblk->numbps - 1) - 4) && (passtype < 2) && (cblksty & J2K_CCP_CBLKSTY_LAZY)) ? T1_TYPE_RAW : T1_TYPE_MQ;
+		/* FIXME: slviewer gets here with a null pointer. Why? Partially downloaded and/or corrupt textures? */
+		if(seg->data == NULL){
+			continue;
+		}
 		if (type == T1_TYPE_RAW) {
-			raw_init_dec(raw, seg->data, seg->len);
+			raw_init_dec(raw, (*seg->data) + seg->dataindex, seg->len);
 		} else {
-			mqc_init_dec(mqc, seg->data, seg->len);
+			mqc_init_dec(mqc, (*seg->data) + seg->dataindex, seg->len);
 		}
 		
 		for (passno = 0; passno < seg->numpasses; ++passno) {
 			switch (passtype) {
 				case 0:
-					t1_dec_sigpass(t1, bpno+1, orient, type, cblksty);
+					if (type == T1_TYPE_RAW) {
+						t1_dec_sigpass_raw(t1, bpno+1, orient, cblksty);
+					} else {
+						if (cblksty & J2K_CCP_CBLKSTY_VSC) {
+							t1_dec_sigpass_mqc_vsc(t1, bpno+1, orient);
+						} else {
+							t1_dec_sigpass_mqc(t1, bpno+1, orient);
+						}
+					}
 					break;
 				case 1:
-					t1_dec_refpass(t1, bpno+1, type, cblksty);
+					if (type == T1_TYPE_RAW) {
+						t1_dec_refpass_raw(t1, bpno+1, cblksty);
+					} else {
+						if (cblksty & J2K_CCP_CBLKSTY_VSC) {
+							t1_dec_refpass_mqc_vsc(t1, bpno+1);
+						} else {
+							t1_dec_refpass_mqc(t1, bpno+1);
+						}
+					}
 					break;
 				case 2:
 					t1_dec_clnpass(t1, bpno+1, orient, cblksty);
@@ -1003,19 +1366,19 @@ static void t1_decode_cblk(
 /* ----------------------------------------------------------------------- */
 
 opj_t1_t* t1_create(opj_common_ptr cinfo) {
-	opj_t1_t *t1 = (opj_t1_t*) malloc(sizeof(opj_t1_t));
+	opj_t1_t *t1 = (opj_t1_t*) opj_malloc(sizeof(opj_t1_t));
 	if(!t1)
 		return NULL;
 
-		t1->cinfo = cinfo;
-		/* create MQC and RAW handles */
-		t1->mqc = mqc_create();
-		t1->raw = raw_create();
+	t1->cinfo = cinfo;
+	/* create MQC and RAW handles */
+	t1->mqc = mqc_create();
+	t1->raw = raw_create();
 
-	t1->datasize=0;
 	t1->data=NULL;
-	t1->flagssize=0;
 	t1->flags=NULL;
+	t1->datasize=0;
+	t1->flagssize=0;
 
 	return t1;
 }
@@ -1025,9 +1388,9 @@ void t1_destroy(opj_t1_t *t1) {
 		/* destroy MQC and RAW handles */
 		mqc_destroy(t1->mqc);
 		raw_destroy(t1->raw);
-		free(t1->data);
-		free(t1->flags);
-		free(t1);
+		opj_aligned_free(t1->data);
+		opj_aligned_free(t1->flags);
+		opj_free(t1);
 	}
 }
 
@@ -1041,23 +1404,30 @@ void t1_encode_cblks(
 	tile->distotile = 0;		/* fixed_quality */
 
 	for (compno = 0; compno < tile->numcomps; ++compno) {
-		opj_tcd_tilecomp_t *tilec = &tile->comps[compno];
+		opj_tcd_tilecomp_t* tilec = &tile->comps[compno];
+		opj_tccp_t* tccp = &tcp->tccps[compno];
+		int tile_w = tilec->x1 - tilec->x0;
 
 		for (resno = 0; resno < tilec->numresolutions; ++resno) {
 			opj_tcd_resolution_t *res = &tilec->resolutions[resno];
 
 			for (bandno = 0; bandno < res->numbands; ++bandno) {
-				opj_tcd_band_t *band = &res->bands[bandno];
+				opj_tcd_band_t* restrict band = &res->bands[bandno];
+        int bandconst = 8192 * 8192 / ((int) floor(band->stepsize * 8192));
 
 				for (precno = 0; precno < res->pw * res->ph; ++precno) {
 					opj_tcd_precinct_t *prc = &band->precincts[precno];
 
 					for (cblkno = 0; cblkno < prc->cw * prc->ch; ++cblkno) {
-						int x, y, w, i, j, orient;
-						opj_tcd_cblk_t *cblk = &prc->cblks[cblkno];
+						opj_tcd_cblk_enc_t* cblk = &prc->cblks.enc[cblkno];
+						int* restrict datap;
+						int* restrict tiledp;
+						int cblk_w;
+						int cblk_h;
+						int i, j;
 
-						x = cblk->x0 - band->x0;
-						y = cblk->y0 - band->y0;
+						int x = cblk->x0 - band->x0;
+						int y = cblk->y0 - band->y0;
 						if (band->bandno & 1) {
 							opj_tcd_resolution_t *pres = &tilec->resolutions[resno - 1];
 							x += pres->x1 - pres->x0;
@@ -1067,46 +1437,49 @@ void t1_encode_cblks(
 							y += pres->y1 - pres->y0;
 						}
 
-						allocate_buffers(
-								t1,
-								cblk->x1 - cblk->x0,
-								cblk->y1 - cblk->y0);
-
-						w = tilec->x1 - tilec->x0;
-						if (tcp->tccps[compno].qmfbid == 1) {
-							for (j = 0; j < t1->h; ++j) {
-								for (i = 0; i < t1->w; ++i) {
-									t1->data[(j * t1->w) + i] =
-										tilec->data[(x + i) + (y + j) * w] << T1_NMSEDEC_FRACBITS;
-								}
-							}
-						} else {		/* if (tcp->tccps[compno].qmfbid == 0) */
-							for (j = 0; j < t1->h; ++j) {
-								for (i = 0; i < t1->w; ++i) {
-									t1->data[(j * t1->w) + i] = 
-										fix_mul(
-										tilec->data[x + i + (y + j) * w],
-										8192 * 8192 / ((int) floor(band->stepsize * 8192))) >> (11 - T1_NMSEDEC_FRACBITS);
-								}
-							}
+						if(!allocate_buffers(
+									t1,
+									cblk->x1 - cblk->x0,
+									cblk->y1 - cblk->y0))
+						{
+							return;
 						}
-						orient = band->bandno;	/* FIXME */
-						if (orient == 2) {
-							orient = 1;
-						} else if (orient == 1) {
-							orient = 2;
+
+						datap=t1->data;
+						cblk_w = t1->w;
+						cblk_h = t1->h;
+
+						tiledp=&tilec->data[(y * tile_w) + x];
+						if (tccp->qmfbid == 1) {
+							for (j = 0; j < cblk_h; ++j) {
+								for (i = 0; i < cblk_w; ++i) {
+									int tmp = tiledp[(j * tile_w) + i];
+									datap[(j * cblk_w) + i] = tmp << T1_NMSEDEC_FRACBITS;
+								}
+							}
+						} else {		/* if (tccp->qmfbid == 0) */
+							for (j = 0; j < cblk_h; ++j) {
+								for (i = 0; i < cblk_w; ++i) {
+									int tmp = tiledp[(j * tile_w) + i];
+									datap[(j * cblk_w) + i] =
+										fix_mul(
+										tmp,
+										bandconst) >> (11 - T1_NMSEDEC_FRACBITS);
+								}
+							}
 						}
 
 						t1_encode_cblk(
 								t1,
 								cblk,
-								orient,
+								band->bandno,
 								compno,
 								tilec->numresolutions - 1 - resno,
-								tcp->tccps[compno].qmfbid,
+								tccp->qmfbid,
 								band->stepsize,
-								tcp->tccps[compno].cblksty,
+								tccp->cblksty,
 								tile->numcomps,
+								tcp->mct,
 								tile);
 
 					} /* cblkno */
@@ -1117,94 +1490,93 @@ void t1_encode_cblks(
 }
 
 void t1_decode_cblks(
-		opj_t1_t *t1,
-		opj_tcd_tile_t *tile,
-		opj_tcp_t *tcp)
+		opj_t1_t* t1,
+		opj_tcd_tilecomp_t* tilec,
+		opj_tccp_t* tccp)
 {
-	int compno, resno, bandno, precno, cblkno;
+	int resno, bandno, precno, cblkno;
 
-	for (compno = 0; compno < tile->numcomps; ++compno) {
-		opj_tcd_tilecomp_t *tilec = &tile->comps[compno];
+	int tile_w = tilec->x1 - tilec->x0;
 
-		for (resno = 0; resno < tilec->numresolutions; ++resno) {
-			opj_tcd_resolution_t *res = &tilec->resolutions[resno];
+	for (resno = 0; resno < tilec->numresolutions; ++resno) {
+		opj_tcd_resolution_t* res = &tilec->resolutions[resno];
 
-			for (bandno = 0; bandno < res->numbands; ++bandno) {
-				opj_tcd_band_t *band = &res->bands[bandno];
+		for (bandno = 0; bandno < res->numbands; ++bandno) {
+			opj_tcd_band_t* restrict band = &res->bands[bandno];
 
-				for (precno = 0; precno < res->pw * res->ph; ++precno) {
-					opj_tcd_precinct_t *prc = &band->precincts[precno];
+			for (precno = 0; precno < res->pw * res->ph; ++precno) {
+				opj_tcd_precinct_t* precinct = &band->precincts[precno];
 
-					for (cblkno = 0; cblkno < prc->cw * prc->ch; ++cblkno) {
-						int x, y, w, i, j, orient, cblk_w, cblk_h;
-						opj_tcd_cblk_t *cblk = &prc->cblks[cblkno];
+				for (cblkno = 0; cblkno < precinct->cw * precinct->ch; ++cblkno) {
+					opj_tcd_cblk_dec_t* cblk = &precinct->cblks.dec[cblkno];
+					int* restrict datap;
+					int cblk_w, cblk_h;
+					int x, y;
+					int i, j;
 
-						orient = band->bandno;	/* FIXME */
-						if (orient == 2) {
-							orient = 1;
-						} else if (orient == 1) {
-							orient = 2;
-						}
-						
-						t1_decode_cblk(
-								t1,
-								cblk,
-								orient,
-								tcp->tccps[compno].roishift,
-								tcp->tccps[compno].cblksty);
+					t1_decode_cblk(
+							t1,
+							cblk,
+							band->bandno,
+							tccp->roishift,
+							tccp->cblksty);
 
-						x = cblk->x0 - band->x0;
-						y = cblk->y0 - band->y0;
-						if (band->bandno & 1) {
-							opj_tcd_resolution_t *pres = &tilec->resolutions[resno - 1];
-							x += pres->x1 - pres->x0;
-						}
-						if (band->bandno & 2) {
-							opj_tcd_resolution_t *pres = &tilec->resolutions[resno - 1];
-							y += pres->y1 - pres->y0;
-						}
+					x = cblk->x0 - band->x0;
+					y = cblk->y0 - band->y0;
+					if (band->bandno & 1) {
+						opj_tcd_resolution_t* pres = &tilec->resolutions[resno - 1];
+						x += pres->x1 - pres->x0;
+					}
+					if (band->bandno & 2) {
+						opj_tcd_resolution_t* pres = &tilec->resolutions[resno - 1];
+						y += pres->y1 - pres->y0;
+					}
 
-						cblk_w = cblk->x1 - cblk->x0;
-						cblk_h = cblk->y1 - cblk->y0;
+					datap=t1->data;
+					cblk_w = t1->w;
+					cblk_h = t1->h;
 
-						if (tcp->tccps[compno].roishift) {
-							int thresh = 1 << tcp->tccps[compno].roishift;
-							for (j = 0; j < cblk_h; ++j) {
-								for (i = 0; i < cblk_w; ++i) {
-									int val = t1->data[(j * t1->w) + i];
-									int mag = int_abs(val);
-									if (mag >= thresh) {
-										mag >>= tcp->tccps[compno].roishift;
-										t1->data[(j * t1->w) + i] = val < 0 ? -mag : mag;
-									}
+					if (tccp->roishift) {
+						int thresh = 1 << tccp->roishift;
+						for (j = 0; j < cblk_h; ++j) {
+							for (i = 0; i < cblk_w; ++i) {
+								int val = datap[(j * cblk_w) + i];
+								int mag = abs(val);
+								if (mag >= thresh) {
+									mag >>= tccp->roishift;
+									datap[(j * cblk_w) + i] = val < 0 ? -mag : mag;
 								}
 							}
 						}
-						
-						w = tilec->x1 - tilec->x0;
-						if (tcp->tccps[compno].qmfbid == 1) {
-							for (j = 0; j < cblk_h; ++j) {
-								for (i = 0; i < cblk_w; ++i) {
-									tilec->data[x + i + (y + j) * w] = t1->data[(j * t1->w) + i]/2;
-								}
-							}
-						} else {		/* if (tcp->tccps[compno].qmfbid == 0) */
-							for (j = 0; j < cblk_h; ++j) {
-								for (i = 0; i < cblk_w; ++i) {
-									if (t1->data[(j * t1->w) + i] >> 1 == 0) {
-										tilec->data[x + i + (y + j) * w] = 0;
-									} else {
-										double tmp = (double)(t1->data[(j * t1->w) + i] * band->stepsize * 4096.0);
-										int tmp2 = ((int) (floor(fabs(tmp)))) + ((int) floor(fabs(tmp*2))%2);									
-										tilec->data[x + i + (y + j) * w] = ((tmp<0)?-tmp2:tmp2);
-									}
-								}
+					}
+
+					if (tccp->qmfbid == 1) {
+						int* restrict tiledp = &tilec->data[(y * tile_w) + x];
+						for (j = 0; j < cblk_h; ++j) {
+							for (i = 0; i < cblk_w; ++i) {
+								int tmp = datap[(j * cblk_w) + i];
+								((int*)tiledp)[(j * tile_w) + i] = tmp / 2;
 							}
 						}
-					} /* cblkno */
-				} /* precno */
-			} /* bandno */
-		} /* resno */
-	} /* compno */
+					} else {		/* if (tccp->qmfbid == 0) */
+						float* restrict tiledp = (float*) &tilec->data[(y * tile_w) + x];
+						for (j = 0; j < cblk_h; ++j) {
+							float* restrict tiledp2 = tiledp;
+							for (i = 0; i < cblk_w; ++i) {
+								float tmp = *datap * band->stepsize;
+								*tiledp2 = tmp;
+								datap++;
+								tiledp2++;
+							}
+							tiledp += tile_w;
+						}
+					}
+					opj_free(cblk->data);
+					opj_free(cblk->segs);
+				} /* cblkno */
+				opj_free(precinct->cblks.dec);
+			} /* precno */
+		} /* bandno */
+	} /* resno */
 }
 
